import io
import sys
import time
import urllib.request
from dataclasses import dataclass, field
from threading import Timer
from typing import Any, Dict, List, Optional, Union, cast

import pandas as pd
import redshift_connector
import snowflake.connector

if sys.version_info.minor != 6:
    from domino_data.data_sources import DatasourceConfig, SnowflakeConfig, _Object

from .domino_api import AppInfo, IDominoApi
from .settings import settings


class Result:
    def to_pandas(self):
        pass


class SnowflakeResult(Result):
    def __init__(self, cs):
        self.cs = cs

    def to_pandas(self):
        try:
            return self.cs.fetch_pandas_all()
        except snowflake.connector.errors.NotSupportedError:
            names = [x.name for x in self.cs.description]
            return pd.DataFrame(self.cs.fetchall(), columns=names)


class DataSource:
    override: Optional[DatasourceConfig] = None

    def query(self, q):
        pass

    def update(self, config: DatasourceConfig):
        self.override = config


class SnowflakeDataSource(DataSource):
    def __init__(self, cs):
        self.cs = cs
        self.datasource_type = "SnowflakeConfig"
        self.config = {}

    def query(self, q):
        if self.override:
            override = cast(SnowflakeConfig, self.override)
            self.cs.execute(f"USE SCHEMA {override.database}.{override.schema}")
        self.cs.execute(q)
        return SnowflakeResult(self.cs)


class RedshiftResult(Result):
    def __init__(self, df):
        self.df = df

    def to_pandas(self):
        return self.df


class RedshiftDataSource(DataSource):
    def __init__(self, cs):
        self.cs = cs
        self.datasource_type = "RedshiftConfig"
        self.config = {"database": "dev"}

    def query(self, q):
        try:
            self.cs.execute(q)
            return RedshiftResult(self.cs.fetch_dataframe() if self.cs.redshift_rowcount > -1 else None)
        finally:
            self.cs.execute("ABORT")


class S3DataSource(DataSource):
    delay = 3.0

    def __init__(self):
        self.datasource_type = "S3Config"

    def list_objects(self):
        time.sleep(self.delay)
        return [
            _Object(
                datasource=self,  # type: ignore
                key=key,
            )
            for key in ["1/foo.txt", "titanic.csv"]
        ]

    def download_fileobj(self, key: str, fileobj):
        with urllib.request.urlopen("https://vve589t3tspu.s3.us-west-2.amazonaws.com/titanic.csv") as f:
            fileobj.write(f.read())

    def get(self, key: str):
        buffer = io.BytesIO()
        self.download_fileobj(key, buffer)
        return buffer.getvalue()


class DataSourceClient:
    def get_datasource(self, name):
        if "snowflake" in name.lower():
            ctx = snowflake.connector.connect(user=settings.snowflake_user, password=settings.snowflake_password, account=settings.snowflake_account)
            cs = ctx.cursor()
            return SnowflakeDataSource(cs)
        if "redshift" in name.lower():
            conn = redshift_connector.connect(
                host=settings.redshift_host,
                database="dev",
                user=settings.redshift_user,
                password=settings.redshift_password,
            )
            return RedshiftDataSource(conn.cursor())
        if "s3" in name.lower():
            return S3DataSource()


@dataclass()
class MockDominoApi(IDominoApi):
    delay: float = 3.0
    data_list: Union[List, Exception] = field(
        default_factory=lambda: [
            {"name": "mario_test_snowflake", "dataSourceType": "SnowflakeConfig"},
            {"name": "mario_test_redshift", "dataSourceType": "RedshiftConfig"},
            {"name": "mario_test_big_query", "dataSourceType": "BigQueryConfig"},
            {"name": "mario_test_s3", "dataSourceType": "S3Config"},
            {"name": "mario_na", "dataSourceType": "NaConfig"},
        ]
    )
    data_sources: Union[List, Exception] = field(
        default_factory=lambda: ["mario_test_snowflake", "mario_test_redshift", "mario_test_big_query", "mario_test_s3"]
    )
    sync_return: Union[bool, Exception] = True
    app_status_delay = 0.1
    app_status: Union[str, Exception, None] = "Running"  # "Pending", "Preparing", "Running", "Stopped" or "Failed"
    app_publish_end_state: Union[str, Exception] = "Running"
    app_unpublish_end_state: Union[str, Exception] = "Stopped"
    app_id: Union[Optional[str], Exception] = "123"
    app_info: List[AppInfo] = field(
        default_factory=lambda: [
            {
                "created": "2022-10-17T11:48:58.606Z",
                "lastUpdated": "2022-10-17T11:48:58.606Z",
                "runningCommitId": "fe15cf09ac8fd55405141bbd81a8276b07d263a8",
                "openUrl": "some/url",
                "latestAppVersionId": "some/version",
                "id": "some/id",
                "hardwareTierId": "small-k8s",
                "status": "Running",
            }
        ]
    )
    current_environment_id: str = "632b014bde659c33feb8f748"
    default_environment_id: str = current_environment_id

    def get_datasource_list(self):
        time.sleep(self.delay)
        return self._return_or_raise(self.data_list)

    def get_datasource_names(self):
        time.sleep(self.delay)
        return self._return_or_raise(self.data_sources)

    def sync(self, commit_message: str):
        time.sleep(self.delay)
        return self._return_or_raise(self.sync_return)

    @staticmethod
    def _return_or_raise(value):
        if isinstance(value, Exception):
            raise value
        return value

    def get_app_status(self):
        time.sleep(self.app_status_delay)
        return self._return_or_raise(self.app_status)

    def app_publish(self, hardwareTierId=None):
        time.sleep(self.app_status_delay)

        self._return_or_raise(self.app_publish_end_state)
        self.app_status = "Pending"

        def set_preparing():
            self.app_status = "Preparing"

        Timer(self.delay, set_preparing).start()

        end_state = self.app_publish_end_state

        def set_end_state():
            self.app_status = end_state

        Timer(self.delay * 2, set_end_state).start()

    def app_unpublish(self):
        time.sleep(self.app_status_delay)

        self._return_or_raise(self.app_unpublish_end_state)

        end_state = self.app_unpublish_end_state

        def set_end_state():
            self.app_status = end_state

        Timer(self.delay, set_end_state).start()

    def get_app_id(self):
        return self._return_or_raise(self.app_id)

    def get_app_info(self) -> List[AppInfo]:
        time.sleep(self.delay)
        return self.app_info

    def get_user_info(self) -> Optional[dict]:
        return {
            "firstName": "John",
            "lastName": "Smith",
            "fullName": "John Smith",
            "userName": "johnsmith",
            "email": "johnsmith@example.com",
            "avatarUrl": "",
            "id": "1234567890b",
        }

    def get_domino_version(self) -> str:
        return "0.0"

    def sync_files(self, commit_message: str) -> Dict[str, Any]:
        return {"succeeded": True, "message": "Synchronizing files ..."}

    def get_useable_environments(self):
        time.sleep(self.delay)
        return {
            "currentlySelectedEnvironment": {
                "id": self.default_environment_id,
                "v2EnvironmentDetails": {
                    "latestRevision": 10,
                    "latestRevisionUrl": "/environments/revisions/6349a8044335d417d82b7036",
                    "latestRevisionStatus": "Succeeded",
                    "selectedRevision": 9,
                    "selectedRevisionUrl": "/environments/revisions/63487b044325d417d82b6d20",
                },
                "supportedClusters": [],
            },
            "environments": [
                {
                    "id": "632b014bde659c33feb8f748",
                    "archived": False,
                    "name": "Domino Standard Environment 5.2 + DCA + AutoML",
                    "version": 2,
                    "visibility": "Private",
                    "owner": {"id": "userid1", "username": "user1"},
                    "supportedClusters": [],
                },
                {
                    "id": "628560094906b439238b950f",
                    "archived": False,
                    "name": "PyCharm",
                    "version": 2,
                    "visibility": "Private",
                    "owner": {"id": "userid2", "username": "user2"},
                    "supportedClusters": [],
                },
                {
                    "id": "63582e044335d417d82b7c3b",
                    "archived": False,
                    "name": "Domino Standard 5.2 + DCA (0.3.1)",
                    "version": 2,
                    "visibility": "Private",
                    "owner": {"id": "userid3", "username": "user2"},
                    "supportedClusters": [],
                },
            ],
        }

    def get_current_environment_id(self) -> str:
        time.sleep(self.delay)
        return self.current_environment_id



from typing import Any, Callable, Dict, List, Optional, cast

import pandas as pd
import plotly.graph_objs
import reacton.ipyvuetify as v
import solara

from domino_code_assist import app, css, util

from .assistant import notebook
from .transform.column_description import ColumnDescription


@solara.component
def DataFrame(df, on_drop_column: Callable[[str], Any] = None, on_filter_value: Callable[[str, int], Any] = None):
    """Extends the solara dataframe component with a custom column description and custom actions."""
    column_menu_open, set_column_menu_open = solara.use_state(cast(Optional[str], None))

    def on_local_on_drop_column(column):
        if on_drop_column:
            on_drop_column(column)

    def on_local_filter_value(column, row_index):
        if on_filter_value:
            on_filter_value(column, row_index)

    column_actions = [solara.ColumnAction(icon="mdi-table-column-remove", name="drop column", on_click=on_local_on_drop_column)]
    cell_actions = [solara.CellAction(icon="mdi-filter", name="Filter values like this", on_click=on_local_filter_value)]
    return solara.DataFrame(
        df,
        column_actions=column_actions,
        cell_actions=cell_actions,
        items_per_page=10,
        scrollable=True,
        on_column_header_hover=set_column_menu_open,
        column_header_info=ColumnDescription(df, column_menu_open),
    )


def _make_grid_panel_viewer(obj):
    def wrap(obj):
        if isinstance(obj, plotly.graph_objs.Figure):
            return solara.FigurePlotly(obj)
        if isinstance(obj, pd.DataFrame):
            return solara.CrossFilterDataFrame(obj, items_per_page=10, scrollable=True)
        elif isinstance(obj, app.ReactiveDf):
            return app.DropDataframe(obj.var_name, obj.label)
        if isinstance(obj, str):
            return MarkdownFromCell(obj)
        return obj

    with v.Card(style_="height: 100%;") as panel:
        v.CardText(style_="height: 100%; max-height: 100%; overflow: auto;", children=[wrap(obj)])

    return panel


@solara.component
def CardGridLayout(layout, resizable=False, draggable=False):
    solara.provide_cross_filter()

    layout_without_item, set_layout_without_item = solara.use_state(cast(List[Dict], []))

    layout_no_item = [
        {
            **util.remove_keys(item, ["item"]),
            "i": index,
        }
        for index, item in enumerate(layout)
    ]

    solara.use_memo(lambda: set_layout_without_item(layout_no_item), [layout_no_item])

    items = [_make_grid_panel_viewer(layout_item["item"]) for layout_item in layout]

    with solara.Div(class_="domino-plotly-auto-height domino-card-grid") as main:
        css.Css()
        solara.GridLayout(items=items, grid_layout=layout_without_item, resizable=resizable, draggable=draggable, on_grid_layout=set_layout_without_item)

    return main


@solara.component
def MarkdownFromCell(cell_id: str):
    if notebook.markdown_cells and cell_id in notebook.markdown_cells:
        solara.Markdown(notebook.markdown_cells[cell_id])
    else:
        solara.Text(f'No markdown cell with id "{cell_id}" found.')


# for backwards compatibility
DataTable = DataFrame



from typing import Callable, List, Optional

import ipyvuetify as v
import reacton
import traitlets


class DfSelectWidget(v.VuetifyTemplate):
    template_file = (__file__, "df_select.vue")

    label = traitlets.Unicode("").tag(sync=True)
    items = traitlets.List().tag(sync=True)
    value = traitlets.Unicode(allow_none=True).tag(sync=True)
    hide_details = traitlets.Any(False).tag(sync=True)
    error_messages = traitlets.List(allow_none=True).tag(sync=True)
    messages = traitlets.List(allow_none=True).tag(sync=True)
    error = traitlets.Bool(False).tag(sync=True)


@reacton.component
def DfSelect(
    label: str,
    items: List,
    value: Optional[str],
    on_value: Callable[[Optional[str]], None],
    hide_details: bool = False,
    error_messages: Optional[List[str]] = None,
    messages: Optional[List[str]] = [],
    error: bool = False,
):
    return DfSelectWidget.element(
        label=label, items=items, value=value, on_value=on_value, hide_details=hide_details, error_messages=error_messages, messages=messages, error=error
    )



import inspect
import types
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple, Union

import plotly.express as px
import plotly.graph_objs._figure
import plotly.io as pio
import reacton
import reacton.ipyvuetify as v
import solara
import solara as sol
from solara.lab import Ref

from domino_code_assist.assistant import mixpanel

from .state import Viz, viz

basic_col_args = (
    "x",
    "y",
    "locations",
    "color",
    "symbol",
    "r",
    "theta",
    "lat",
    "lon",
    "z",
    "a",
    "b",
    "c",
    "size",
)
extra_col_args = (
    "line_group",
    "symbol",
    "pattern_shape",
    "line_dash",
    "names",
    "values",
    "parents",
    "text",
    "hover_name",
    "hover_data",
    "facet_row",
    "facet_col",
    "base",
)

col_list_args = ("hover_data",)

basic_plots = (
    "area",
    "bar",
    "histogram",
    "line",
    "line_3d",
    "scatter",
    "scatter_3d",
    "scatter_geo",
    "choropleth",
)

translations = {
    "area": "Area",
    "bar": "Bar",
    "histogram": "Histogram",
    "line": "Line",
    "line_3d": "Line 3D",
    "scatter": "Scatter",
    "scatter_3d": "Scatter 3D",
    "scatter_geo": "Scatter Map",
    "choropleth": "Choropleth Map",
    "x": "X-axis",
    "y": "Y-axis",
    "z": "Z-axis",
}


def un_dash(value):
    return " ".join([el.capitalize() for el in value.split("_")])


def get_args(name):
    return list(inspect.signature(getattr(px, name)).parameters.keys())


def get_col_args(name):
    return [a for a in get_args(name) if a in basic_col_args + extra_col_args]


specs = {
    n: {
        "name": n,
        "col_args": get_col_args(n),
        "args": get_args(n),
    }
    for n in dir(px)
    if isinstance(getattr(px, n), types.FunctionType) and get_args(n) and get_args(n)[0] == "data_frame"
}


arg_types: Dict[str, Dict[str, Any]] = {
    "template": {
        "type": list(pio.templates._templates.keys()),
        "default": "plotly",
        "name": "Theme",
    },
    "orientation": {
        "type": ["h", "v"],
        "default": "v",
        "type_mapping": ["Horizontal", "Vertical"],
        "name": "Orientation",
    },
    "histnorm": {
        "type": ["percent", "probability", "density"],
        "name": "Normalization",
    },
    "histfunc": {
        "type": ["count", "sum", "avg", "min", "max"],
        "default": "count",
        "name": "Aggregate Function",
    },
    "log_x": {"type": bool, "default": False, "description": "If True, the x-axis is log-scaled in cartesian coordinates", "name": "Logarithmic X-axis"},
    "log_y": {
        "type": bool,
        "default": False,
        "description": "If True, the x-axis is log-scaled in cartesian coordinates",
        "name": "Logarithmic Y-axis",
    },
    "color_continuous_scale": {
        "type": [x for x in dir(px.colors.sequential) if not x.startswith("_")],
        "default": "Plasma",
        "name": "Color Scale",
    },
}

plot_types = basic_plots


def nop(*_):  # pragma: no cover
    return


plot_type_items = [dict(value=t, text=translations[t]) for t in plot_types]


def track_set_attribute(col_arg):
    mixpanel.api.track_with_defaults(
        "interaction",
        {
            "section": "visualizations",
            "type": "set attribute",
            "attribute": col_arg,
        },
    )


def generate_code(viz: Viz):
    args = {k: v for k, v in viz.col_args.items() if k in specs[viz.plot_type]["args"]}  # type: ignore

    def transform_arg(k, value):
        if k == "color_continuous_scale":
            return f"px.colors.sequential.{value}"
        if type(value) == str:
            return f'"{value}"'
        return value

    arg_str = ", ".join(
        [f"{k}={transform_arg(k, v)}" for k, v in args.items() if v is not None and not (isinstance(v, list) and len(v) == 0)],
    )

    def remove_margins(var_name):
        return f"{var_name}.update_layout(margin=dict(l=0, r=0, t=40 if {var_name}.layout.title.text else 20, b=0))"

    if viz.crossfilter_enabled:
        base_var_name = "_base_" + (viz.plot_var_name or "")
        select_config = ""
        if viz.plot_type == "scatter":
            select_config = f'{base_var_name}.update_layout(dragmode="lasso")\n'
        elif viz.plot_type in ["bar", "histogram"]:
            direction = "v" if "orientation" in viz.col_args and viz.col_args["orientation.value"] == "h" else "h"
            select_config = f'{base_var_name}.update_layout(dragmode="select", selectdirection="{direction}")\n'
        return "".join(
            [
                "from solara.express import CrossFilteredFigurePlotly\n",
                "import plotly.express as px\n\n",
                f"{base_var_name} = px.{viz.plot_type}({viz.df_var_name}, {arg_str})\n",
                select_config,
                remove_margins(base_var_name) + "\n",
                f"{viz.plot_var_name} = " if viz.plot_var_name else "",
                f"CrossFilteredFigurePlotly({base_var_name})\n",
                f"\n{viz.plot_var_name}" if viz.plot_var_name else "",
            ]
        )
    return "".join(
        [
            "import plotly.express as px\n\n",
            f"{viz.plot_var_name} = " if viz.plot_var_name else "",
            f"px.{viz.plot_type}({viz.df_var_name}, {arg_str})\n",
            remove_margins(viz.plot_var_name),
            f"\n{viz.plot_var_name}" if viz.plot_var_name else "",
        ]
    )


@reacton.component
def DimensionSelects(columns, for_cols: Union[List[str], Tuple[str, ...]]):
    viz.use()
    assert viz.value.plot_type is not None
    with v.Row() as main:
        for col_arg in specs[viz.value.plot_type]["col_args"]:

            def set_attribute(value, col_arg=col_arg):
                track_set_attribute(col_arg)
                Ref(viz.fields.col_args[col_arg]).value = value

            if col_arg not in for_cols:
                continue
            with v.Col(sm=2):
                with solara.Div().key(str(col_arg)):
                    v.Select(
                        label=translations.get(col_arg) or un_dash(col_arg),
                        items=columns,
                        clearable=True,
                        multiple=col_arg in col_list_args,
                        v_model=viz.value.col_args.get(col_arg),
                        on_v_model=set_attribute,
                    ).meta(ref=f"dimension_{col_arg}")
    return main


@reacton.component
def OptionsPanel():
    viz.use()
    assert viz.value.plot_type is not None
    with v.Row() as main:
        for arg in specs[viz.value.plot_type]["args"]:

            def set_attribute(value, col_arg=arg):
                track_set_attribute(col_arg)
                Ref(viz.fields.col_args[col_arg]).value = value

            if arg not in arg_types:
                continue
            with v.Col(sm=2):
                with solara.Div():
                    arg_spec = arg_types[arg]
                    if arg_spec.get("type") == bool:
                        v.Switch(
                            label=arg_spec["name"],
                            v_model=viz.value.col_args.get(arg),
                            on_v_model=set_attribute,
                        ).meta(ref=f"dimension_{arg}")
                    elif type(arg_spec["type"]) in (list, tuple):
                        v.Select(
                            label=arg_spec["name"],
                            items=[dict(value=v, text=t) for v, t in zip(arg_spec["type"], arg_spec["type_mapping"])]
                            if arg_spec.get("type_mapping")
                            else arg_spec["type"],
                            v_model=viz.value.col_args.get(arg),
                            on_v_model=set_attribute,
                        ).meta(ref=f"dimension_{arg}")
    return main


@dataclass(frozen=True)
class FigState:
    fig: Optional[plotly.graph_objs._figure.Figure] = None
    error: Optional[str] = None
    code: Optional[str] = None


@reacton.component
def PlotBuilder(df, columns, min_height="724px"):
    viz.use()
    state, set_state = reacton.use_state(FigState())
    exp_panel, set_exp_panel = reacton.use_state(0)
    auto_preview, set_auto_preview = reacton.use_state(True)
    update_preview, set_update_preview = reacton.use_state(0)
    should_update, set_should_update = reacton.use_state(False)

    def make_fig():
        if not should_update:
            return
        if not Ref(viz.fields.col_args).value or not viz.value.plot_type or not df:
            set_state(FigState())
            set_should_update(False)
            return
        try:
            args = {k: v for k, v in viz.value.col_args.items() if k in specs[viz.value.plot_type]["args"]}
            fig = getattr(px, viz.value.plot_type)(df.get(), **args)
            code = generate_code(viz.value)
            set_state(FigState(fig=fig, code=code))
        except Exception as e:
            set_state(FigState(error=str(e)))
        set_should_update(False)

    reacton.use_memo(lambda: set_should_update(True), [viz.value, df])
    reacton.use_effect(make_fig, [auto_preview and should_update, update_preview])

    with v.Sheet(style_=min_height and f"min-height: {min_height}") as main:

        def on_plot_type(plot_type):
            mixpanel.api.track_with_defaults(
                "interaction",
                {
                    "section": "visualizations",
                    "type": "selected plot type",
                    "plot_type": plot_type,
                },
            )
            Ref(viz.fields.plot_type).value = plot_type

        def on_crossfilter(value):
            mixpanel.api.track_with_defaults(
                "interaction",
                {
                    "section": "visualizations",
                    "type": "set crossfilter",
                },
            )
            Ref(viz.fields.crossfilter_enabled).value = value

        v.Select(
            label="Plot Type",
            items=plot_type_items,
            v_model=viz.value.plot_type,
            on_v_model=on_plot_type,
        ).meta(ref="plot_type")
        with v.Row():
            with v.Col(sm=2):
                v.Switch(label="Enable crossfilter", v_model=viz.value.crossfilter_enabled, on_v_model=on_crossfilter).meta(ref="crossfilter")
            with v.Col(sm=8, align_self="center"):
                controllers = ["scatter", "bar", "histogram"]
                v.Html(
                    tag="div",
                    children=(
                        "Cross-filtering allows you to interactively explore the relationship between multiple variables in a dataset by "
                        "filtering the data in one plot based on the selection made in another plot or widget."
                    ),
                )
                if viz.value.plot_type and viz.value.plot_type not in controllers and viz.value.crossfilter_enabled:
                    v.Html(
                        tag="span",
                        children="Note: this plot type can only view cross-filtered data, not select it.",
                        style_="padding: 4px; background-color: #ECE63D",
                    ).key("crossfilter info").meta(ref="crossfilter_info")

        if viz.value.plot_type:
            with v.Sheet():
                with v.ExpansionPanels(v_model=exp_panel, on_v_model=set_exp_panel):
                    with v.ExpansionPanel():
                        v.ExpansionPanelHeader(children=["Dimensions"])
                        with v.ExpansionPanelContent():
                            DimensionSelects(columns, basic_col_args)
                    with v.ExpansionPanel():
                        v.ExpansionPanelHeader(children=["Extra Dimensions"])
                        with v.ExpansionPanelContent():
                            DimensionSelects(columns, extra_col_args)
                    with v.ExpansionPanel():
                        v.ExpansionPanelHeader(children=["Options"])
                        with v.ExpansionPanelContent():
                            OptionsPanel()
            with solara.Div(style_="display: flex"):
                v.Switch(label="Auto preview", v_model=auto_preview, on_v_model=set_auto_preview).meta(ref="auto_preview")
                if not auto_preview:
                    sol.Button(
                        "Update preview",
                        icon_name="mdi-refresh-circle",
                        on_click=lambda: set_update_preview(update_preview + 1),
                        disabled=not should_update,
                        color="primary",
                        class_=" ma-3",
                    ).meta(ref="update_preview")
            if state.error:
                solara.Div(children=["plot error", state.error]).meta(ref="error")
            elif state.fig:
                solara.FigurePlotly(state.fig, dependencies=state).meta(ref="figure")
                from solara.components.code_highlight_css import CodeHighlightCss

                CodeHighlightCss()
                from domino_code_assist.code import Code

                Code(
                    code_chunks=[state.code],
                    on_event=lambda a: None,
                    error=None,
                ).meta(ref="code")

    return main



from playwright.sync_api import Page

from domino_code_assist.playwright.app import AppHelper
from domino_code_assist.playwright.load_data import LoadDataHelper
from domino_code_assist.playwright.transform import TransformHelper


class AssistentHelper:
    def __init__(self, page: Page) -> None:
        self.page = page

    @property
    def initializer(self):
        return self.page.locator('[aria-label="Domino Code Assist"]')

    @property
    def initialized_text(self):
        return self.page.locator("text=Domino Code Assist initialized")

    @property
    def domino_logo(self):
        return self.page.locator(".domino_code_assist-assistant-menu")

    @property
    def load_data(self):
        return LoadDataHelper(self.page)

    @property
    def app(self):
        return AppHelper(self.page)

    @property
    def transform(self):
        return TransformHelper(self.page)

    @property
    def insert_code(self):
        return self.page.locator("text=Insert Code")



from functools import partial
from typing import Dict, List, Optional, Union, cast

import reacton.ipyvuetify as v
import solara

from domino_code_assist.automl import flaml_api

auto = solara.reactive(cast(bool, True))
learner_params = solara.reactive(cast(Dict[str, Dict], {}))
selected_learners = solara.reactive(cast(Optional[List[str]], []))


@solara.component
def Hyperparameters(task: str, auto=auto, learner_params=learner_params, selected_learners=selected_learners):
    all_learners = flaml_api.get_estimator_ids_for_task(task)

    def compute_select_all():
        if len(selected_learners.value or []) == 0:
            return False
        elif len(selected_learners.value) == len(all_learners):
            return True
        else:
            return None

    select_all = compute_select_all()

    opened_estimator, set_opened_estimator = solara.use_state(cast(Optional[Dict], None))

    def on_select_learner(learner: str, on: bool):
        if on:
            learners = selected_learners.value + [learner]
        else:
            learners = [sl for sl in selected_learners.value if sl != learner]
        selected_learners.set(learners)

    def on_select_all(on: bool):
        selected_learners.set([*all_learners] if on else [])

    def on_open_params(estimator: Dict):
        params = learner_params.value.get(estimator["id"], {})
        params_selected.value = params.get("selected", [])
        params_lower.value = params.get("lower", {})
        params_upper.value = params.get("upper", {})
        set_opened_estimator(estimator)

    def on_save():
        if opened_estimator:
            learner_params.value = {
                **learner_params.value,
                opened_estimator["id"]: {
                    "selected": params_selected.value,
                    "lower": params_lower.value,
                    "upper": params_upper.value,
                },
            }
            set_opened_estimator(None)

    def get_enabled_hyper_params(estimator: Dict) -> List[str]:
        return learner_params.value.get(estimator["id"], {}).get("selected", [])

    with v.ExpansionPanel(class_="dca-hyperparameters"):
        with v.ExpansionPanelHeader().key("header"):
            with solara.Div(style_="display: flex; flex-direction: row; max-width: 100%; overflow: hidden;"):
                with solara.Div(style_="min-width: 140px"):
                    v.Html(tag="strong", children=["Advanced "])
                with solara.Div():
                    solara.Text("Hyperparameter search space: ", style="font-weight: bold; color: var(--jp-border-color1);")
                    solara.Text("auto" if auto.value else "custom")
        with v.ExpansionPanelContent().key("content"):
            solara.Text("Hyperparameter Search Space")
            v.Switch(v_model=auto.value, on_v_model=auto.set, label="Auto (recommended) Domino will auto-tune hyperparameters")
            with solara.Div(class_=f"dca-hyperparameters-list {'dca-hyperparameters-list-off' if auto.value else ''}"):
                with solara.Div():
                    v.Checkbox(label="Select all", v_model=select_all, on_v_model=on_select_all, indeterminate=select_all is None)
                for estimator in [item for item in flaml_api.hyperparameters if task in item["tasks"]]:
                    with solara.ColumnsResponsive([6, 5, 1], gutters=False, classes=["pa-0"], style="line-height: 2.2"):
                        with solara.Row():
                            v.Checkbox(
                                v_model=estimator["id"] in (selected_learners.value or []),
                                on_v_model=partial(on_select_learner, estimator["id"]),
                                class_="dca-hyperparameters-checkbox",
                                hide_details=True,
                            )
                            solara.Text("Learner: ", style="color: var(--jp-border-color1); margin-left: -12px;")
                            solara.Text(estimator.get("readable_name", estimator["id"]))
                        with solara.Row():
                            solara.Text("Hyper-parameters: ", style="color: var(--jp-border-color1);")
                            solara.Text(
                                f" {len(get_enabled_hyper_params(estimator))}/{len(flaml_api.get_hyper_parameters_for_task(estimator, task))} configured"
                            )
                        with solara.Row():
                            solara.Button(icon_name="mdi-pencil", icon=True, on_click=lambda estimator=estimator: on_open_params(estimator))
        if opened_estimator:
            with v.Dialog(v_model=bool(opened_estimator), on_v_model=lambda on: set_opened_estimator(None) if not on else None, max_width="800px").key(
                "dialog"
            ):
                with v.Card(class_="dca-hyperparameters-dialog"):
                    with v.CardTitle():
                        solara.Text("Edit hyper-parameters for ", style="white-space: pre;")
                        solara.Text(opened_estimator.get("readable_name", opened_estimator["id"]), style="font-weight: bold;")
                    with v.CardText():
                        HyperParametersPanel(task, opened_estimator["id"])
                    with v.CardActions():
                        solara.Button(color="primary", text=True, children=["Save"], on_click=on_save)
                        solara.Button(color="primary", text=True, children=["Cancel"], on_click=lambda: set_opened_estimator(None))


params_selected = solara.reactive(cast(List[str], []))
params_lower = solara.reactive(cast(Dict[str, Union[str, List]], {}))
params_upper = solara.reactive(cast(Dict[str, str], {}))


@solara.component
def HyperParametersPanel(
    task: str,
    estimator_id: str,
    selected=params_selected,
    upper=params_upper,
    lower=params_lower,
):
    estimator = [item for item in flaml_api.hyperparameters if item["id"] == estimator_id][0]

    available_params = flaml_api.get_hyper_parameters_for_task(estimator, task)

    def compute_select_all():
        if len(selected.value) == 0:
            return False
        elif len(selected.value) == len(available_params):
            return True
        else:
            return None

    select_all = compute_select_all()

    def on_select_param(param: str, on: bool):
        if on:
            params = selected.value + [param]
        else:
            params = [sl for sl in selected.value if sl != param]
        selected.set(params)

    def on_select_all(on: bool):
        selected.set(available_params if on else [])

    def on_param(param, prop, value):
        if prop == "lower":
            lower.value = {**lower.value, param: value}
        if prop == "upper":
            upper.value = {**upper.value, param: value}

    with solara.Div():
        with solara.Div():
            v.Checkbox(label="Select all", v_model=select_all, on_v_model=on_select_all, indeterminate=select_all is None)
        for param in available_params:
            with solara.ColumnsResponsive([4, 4, 4], gutters=False, classes=["pa-0"], style="line-height: 2.2"):
                type_ = estimator["hyper-parameters"][param]["type"]  # type: ignore
                with solara.Row(classes=["pt-4"]):
                    v.Checkbox(
                        v_model=param in selected.value,
                        on_v_model=partial(on_select_param, param),
                        class_="dca-hyperparameters-checkbox",
                        hide_details=True,
                    )
                    solara.Text("Parameter: ", style="color: var(--jp-border-color1); margin-left: -12px;")
                    solara.Text(param)
                if isinstance(type_, List):
                    with solara.Row().key("choice1"):
                        v.Select(
                            label="choice",
                            items=[{"text": str(choice), "value": choice} for choice in type_],
                            multiple=True,
                            v_model=lower.value.get(param),
                            on_v_model=partial(on_param, param, "lower"),
                            class_="ml-4",
                        )
                    solara.Row().key("choice2")
                else:
                    with solara.Row().key("other1"):
                        v.TextField(
                            label="lower",
                            v_model=lower.value.get(param),
                            on_v_model=partial(on_param, param, "lower"),
                            class_="ml-4",
                        )
                    with solara.Row().key("other2"):
                        v.TextField(
                            label="upper",
                            v_model=upper.value.get(param),
                            on_v_model=partial(on_param, param, "upper"),
                            class_="ml-4",
                        )


def get_hyper_parameters_string(custom_learners: Dict, custom_learners_enabled: List[str]):
    def to_domain(type_: Union[str, List], lower: Union[str, List], upper: str):
        if type_ == "int":
            return {
                "domain": f"tune.lograndint(lower={lower}, upper={upper})",
            }
        if type_ == "float":
            return {
                "domain": f"tune.loguniform(lower={lower}, upper={upper})",
            }
        if isinstance(type_, List):
            return {
                "domain": f"tune.choice({ [None if e == 'None' else e for e in lower]})",
            }

    def learner(k, learner_dict):
        estimator = [e for e in flaml_api.hyperparameters if e["id"] == k][0]

        def get_type(param) -> Union[str, List]:
            return estimator["hyper-parameters"][param]["type"]  # type: ignore

        return {param: to_domain(get_type(param), learner_dict["lower"][param], learner_dict.get("upper", {}).get(param)) for param in learner_dict["selected"]}

    cv_learners = {k: learner(k, v) for k, v in custom_learners.items() if k in custom_learners_enabled}

    def dict_to_string(d, indent=0):
        output = []
        for key, value in d.items():
            if isinstance(value, dict):
                output.append(f"""{' ' * indent}"{key}": {{""")
                output.append(dict_to_string(value, indent + 4))
                output.append(f"{' ' * indent}}},")
            else:
                output.append(f"""{' ' * indent}"{key}": {value}""")
        return "\n".join(output)

    return "{\n" + dict_to_string(cv_learners, 4) + "\n}"



import dataclasses
import functools
import operator
import os
import re
import subprocess
from pathlib import Path
from typing import Any, Dict, List

from ..settings import settings

EDIT_SNIPPET_PREFIX = "## DCA edit snippet:"


@dataclasses.dataclass(frozen=True)
class Snippet:
    path: Path
    name: str
    description: str
    code: str
    base_path: Path


def get_shared_project_paths():
    return [os.environ[k] for k in os.environ.keys() if re.search(r"^DOMINO_.+_WORKING_DIR$", k)]


def get_shared_dataset_paths():
    datasets_dir = Path(settings.domino_datasets_dir)
    if not datasets_dir.exists():
        return []
    local_dir = Path(f"{settings.domino_datasets_dir}/local")
    return [str(p) for p in datasets_dir.glob("*") if p.is_dir() and p != local_dir]


def get_shared_git_paths():
    path = settings.domino_repos_dir or settings.domino_imported_code_dir
    if not path:
        return []
    return [str(p) for p in Path(path).glob("*") if p.is_dir() and (p / "snippets").is_dir()]


def get_prefix_paths():
    return [str(p) for p in settings.domino_snippet_builtin_dir.glob("*") if p.is_dir()]


def read_snippets(sub: Path, data: str, base_path: Path) -> List[Snippet]:
    return [Snippet(Path(*sub.parts[:-1]), sub.parts[-1], "", data, base_path)]


def read_snippet_file(sub: Path, file: Path, base_path: Path) -> List[Snippet]:
    with open(file, "r") as f:
        content = f.read()

    return read_snippets(sub, content, base_path)


def flatten(nested_list: List):
    return functools.reduce(operator.iconcat, nested_list, [])


def find_snippets_in_path(path: Path) -> List[Snippet]:
    files = path.glob("**/*.py")
    base_path = path
    return flatten([read_snippet_file(file.relative_to(path), file, base_path) for file in files])


def snippet_directory(directory: Path):
    # directories under $prefix/share/domino-code-assist/snippets/<name> will not need a suffix
    if settings.domino_snippet_builtin_dir in directory.parents:
        return directory
    # all other directories will need the "snippets" suffix
    return Path(directory) / "snippets"


def find_snippets(paths: List[str]) -> List[Snippet]:
    paths_ = [snippet_directory(Path(p)) for p in paths]
    paths_ = [p for p in paths_ if p.is_dir()]
    return flatten([find_snippets_in_path(path) for path in paths_])


def snippet_list_to_map(snippets: List[Snippet]):
    snippet_map: Dict[str, Any] = {}

    for snippet in snippets:
        sub_map = snippet_map
        for part in snippet.path.parts:
            if part not in sub_map:
                sub_map[part] = {}
            sub_map = sub_map[part]
        sub_map[snippet.name] = snippet

    return snippet_map


def make_get_items(snippet_map):
    def get_items(path: List[Dict]):
        sub_map = snippet_map
        for part in path[1:]:
            sub_map = sub_map[part["id"]]

        # sort by leaf first, then by name case-insensitive alphabetically
        sorted_items = sorted(sub_map.items(), key=lambda t: ("a" if isinstance(t[1], Snippet) else "b") + t[0].lower())

        return [
            {
                "label": re.sub(".py$", "", k),
                "id": k,
                "leaf": isinstance(v, Snippet),
                "icon": "mdi-card-text-outline" if isinstance(v, Snippet) else "mdi-folder-outline",
            }
            for k, v in sorted_items
        ]

    return get_items


def make_get_content(snippet_map):
    def get_content(path: List[Dict]):
        sub_map = snippet_map
        for part in path[1:]:
            sub_map = sub_map[part["id"]]

        return sub_map

    return get_content


def get_writable_paths():
    git_paths = get_shared_git_paths()

    def is_writable(path):
        return 0 == subprocess.run(["git", "push", "--dry-run", "--force", "--no-verify"], cwd=path, stdout=subprocess.PIPE, stderr=subprocess.PIPE).returncode

    return [{"path": str(p), "writable": is_writable(p)} for p in git_paths] + [{"path": str(settings.domino_working_dir), "writable": True}]


def adjust_path(snippet_map, path: List[Dict]):
    if not path:
        return path
    new_path = path[:1]
    sub_map = snippet_map
    try:
        for part in path[1:]:
            sub_map = sub_map[part["id"]]
            new_path.append(part)
    except KeyError:
        return new_path
    return path


def make_fns():

    snippets = find_snippets(
        [
            *get_shared_project_paths(),
            *get_shared_dataset_paths(),
            *get_shared_git_paths(),
            *get_prefix_paths(),
            settings.domino_working_dir,
        ]
    )

    snippet_map = snippet_list_to_map(snippets)
    return len(snippets), make_get_items(snippet_map), make_get_content(snippet_map), functools.partial(adjust_path, snippet_map)


def save_snippet(code: str):
    if code.startswith(EDIT_SNIPPET_PREFIX):
        marker, rest = code.split("\n", 1)
        path = Path(marker.split(EDIT_SNIPPET_PREFIX)[-1].strip())
        path.parent.mkdir(parents=True, exist_ok=True)
        with open(path, "w") as f:
            f.write(rest)



import os
import sys
import threading
import uuid
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, cast

from typing_extensions import TypedDict

from domino_code_assist import util

from .settings import settings

host = settings.domino_alternative_api_host or settings.domino_api_host

_domino_api = None
_lock = threading.Lock()


class AppInfo(TypedDict):
    created: str
    lastUpdated: str
    runningCommitId: str
    openUrl: str
    latestAppVersionId: str
    id: str
    hardwareTierId: str
    status: str


class LogLine(TypedDict):
    timestamp: int  # e.g. 1672151560992
    logType: str  # stderr/stdout/prepareoutput
    log: str  # single line
    size: int  # e.g. 18


def get_domino_api():
    global _domino_api

    with _lock:
        if not _domino_api:
            api_class = settings.domino_api_class
            if api_class:
                import importlib

                module, class_ = api_class.split(":")
                DominoApiClass = getattr(importlib.import_module(module), class_)
                _domino_api = DominoApiClass()
            elif util.in_dev_mode():
                from .domino_lab_mock import MockDominoApi

                _domino_api = MockDominoApi()
            elif sys.version_info.minor == 6:
                from .py36 import Py36DominoApi

                _domino_api = Py36DominoApi()
            else:
                _domino_api = DominoApi()
    return _domino_api


class IDominoApi(ABC):
    @abstractmethod
    def get_datasource_list(self):
        pass

    @abstractmethod
    def get_datasource_names(self) -> List[str]:
        pass

    @abstractmethod
    def sync(self, commit_message: str):
        pass

    @abstractmethod
    def get_app_status(self) -> Optional[str]:
        pass

    @abstractmethod
    def app_publish(self, hardwareTierId: Optional[str]):
        pass

    @abstractmethod
    def app_unpublish(self):
        pass

    @abstractmethod
    def get_app_id(self) -> Optional[str]:
        pass

    def get_default_hardware_tier_id(self):
        return "small-k8s"

    @abstractmethod
    def get_user_info(self) -> Optional[dict]:
        pass

    @abstractmethod
    def get_domino_version(self) -> str:
        pass

    @abstractmethod
    def sync_files(self, commit_message: str) -> Dict[str, Any]:
        pass

    def get_client_version(self) -> str:
        return "0.0"

    @abstractmethod
    def get_app_info(self) -> List[AppInfo]:
        pass

    def get_app_log(self) -> List[LogLine]:
        return [
            {"timestamp": 1672151702389, "logType": "prepareoutput", "log": "Checking for local changes...", "size": 29},
            {"timestamp": 1672151560992, "logType": "stdout", "log": "Using curl at curl", "size": 18},
            {"timestamp": 1672151561051, "logType": "stdout", "log": "Preparing working directory.", "size": 28},
        ] * 100

    @abstractmethod
    def get_useable_environments(self):
        pass

    @abstractmethod
    def get_current_environment_id(self):
        pass

    def get_experiments(self):
        return [
            {
                "experiment_id": "17",
                "name": "popular-fish-163",
                "artifact_location": "mlflow-artifacts:/mlflow",
                "lifecycle_stage": "active",
                "last_update_time": 1682528255165,
                "creation_time": 1682528255165,
                "tags": [{"key": "mlflow.domino.project", "value": "DCA-latest"}],
            },
            {
                "experiment_id": "16",
                "name": "gaudy-donkey-172",
                "artifact_location": "mlflow-artifacts:/mlflow",
                "lifecycle_stage": "active",
                "last_update_time": 1682514840966,
                "creation_time": 1682514840966,
                "tags": [{"key": "mlflow.domino.project", "value": "DCA-latest"}],
            },
        ]


class DominoApi(IDominoApi):
    def get_datasource_list(self):
        project_id = settings.domino_project_id
        return self._get_domino_client().request_manager.get(f"{host}/v4/datasource/projects/{project_id}").json()

    def get_datasource_names(self):
        return [k["name"] for k in self.get_datasource_list()]

    def sync(self, commit_message: str):
        run_id = settings.domino_run_id
        project_owner = settings.domino_project_owner
        project_name = settings.domino_project_name
        response = self._get_domino_client().request_manager.post(
            f"{host}/u/{project_owner}/{project_name}/run/synchronizeRunWorkingDirectory/{run_id}",
            json={
                "uploadLocalChanges": True,
                "shouldSaveConflicts": False,
                "syncOperationInfo": {"syncOperationId": str(uuid.uuid4()), "syncOperationType": "onlyDfs"},
                "commitMessage": commit_message,
            },
        )
        return response.json()["succeeded"]

    def get_app_status(self):
        domino = self._get_domino_client()
        app_id = domino._app_id
        return domino._Domino__app_get_status(app_id)

    def app_publish(self, hardwareTierId=None):
        self._get_domino_client().app_publish(hardwareTierId=hardwareTierId)

    def app_unpublish(self):
        self._get_domino_client().app_unpublish()

    def get_app_id(self):
        return self._get_domino_client()._app_id

    def get_user_info(self):
        return self._get_domino_client().request_manager.get(f"{host}/v4/users/self").json()

    def get_domino_version(self) -> str:
        return self._get_domino_client()._version

    def sync_files(self, commit_message: str) -> Dict[str, Any]:
        api = self._get_domino_client()
        url_base = api._routes._build_project_url_private_api()
        return api.request_manager.post(
            f"{url_base}/run/synchronizeRunWorkingDirectory/{os.environ['DOMINO_RUN_ID']}",
            json={
                "uploadLocalChanges": True,
                "shouldSaveConflicts": False,
                "syncOperationInfo": {"syncOperationId": "0de13390-7ff2-4df8-a016-0512f227ce65", "syncOperationType": "onlyDfs"},
                "commitMessage": commit_message,
            },
        ).json()

    def get_client_version(self):
        from domino._version import __version__

        return __version__

    def get_app_info(self) -> List[AppInfo]:
        api = self._get_domino_client()
        project_id = api.project_id if hasattr(api, "project_id") else api._project_id
        url_str = api._routes.app_list(project_id)
        return cast(List[AppInfo], api.request_manager.get(url_str).json())

    _domino_client = None
    _lock = threading.Lock()

    def _get_domino_client(self):
        with self._lock:
            if not self._domino_client:
                project_owner = settings.domino_project_owner
                project_name = settings.domino_project_name

                # needed for python 3.6 support
                from domino import Domino

                self._domino_client = (
                    Domino(f"{project_owner}/{project_name}", host=settings.domino_alternative_api_host) if not util.in_dev_mode() else cast(Domino, None)
                )
        return self._domino_client

    def get_app_log(self) -> List[LogLine]:
        app_info = self.get_app_info()
        app_id = app_info[0]["latestAppVersionId"]
        return self._get_domino_client().request_manager.get(f"{host}/v4/modelProducts/apps/versions/{app_id}/realTimeLogs").json()["logContent"]

    def get_useable_environments(self):
        return self._get_domino_client()._routes.useable_environments_list()

    def get_current_environment_id(self):
        project_id = settings.domino_project_id
        run_id = settings.domino_run_id
        return (
            self._get_domino_client()
            .request_manager.get(f"{host}/v4/workspaces/{run_id}/runtimeExecutionDetails?projectId={project_id}")
            .json()["environment"]["environmentId"]
        )

    def get_experiments(self):
        project_id = settings.domino_project_id
        return (
            self._get_domino_client()
            .request_manager.post(
                f"{host}/api/2.0/mlflow/experiments/search", json={"max_results": 3000}, headers={"X-Domino-Mlflow-Data": f'{{ "projectId": "{project_id}" }}'}
            )
            .json()["experiments"]
        )



import json
import os
import textwrap
import traceback
from pathlib import Path
from typing import Any, Callable, Dict, Optional, cast

import ipyvuetify as vy
import ipywidgets
import plotly
import reacton
import reacton.ipyvuetify as v
import solara as sol
import solara.lab
import traitlets
from pandas.core.frame import DataFrame
from solara.components.code_highlight_css import CodeHighlightCss

import domino_code_assist.actions_store as actions_store
import domino_code_assist.assistant.notebook
from domino_code_assist import (
    __version__,
    action,
    app,
    crossfilter_widgets,
    css,
    data,
    sample_project_md,
    util,
)
from domino_code_assist.assistant import drawer, handle_user_install, mixpanel, notebook
from domino_code_assist.code import Code
from domino_code_assist.layout import CardGridLayoutBuilder
from domino_code_assist.serialize import dumps, loads
from domino_code_assist.settings import settings
from domino_code_assist.util import get_vars
from domino_code_assist.viz.state import Viz

from ..action import ActionFilter, ActionUseNbLocals
from ..automl.drawer import AutoMlDrawer
from ..automl.store import MLTask
from ..crossfilter_widgets import WidgetState, _generate_map, to_json
from ..load_data.drawer import LoadDataDrawer
from ..snippets import snippets, snippets_ui
from ..snippets.snippet_drawer import SnippetDrawer
from ..tour.tour import TourWidget
from ..transform.transform_drawer import TransformDrawer
from ..viz import plot_builder
from ..viz.drawer import VisualizationDrawer

server_software = os.environ.get("SERVER_SOFTWARE", "")
if server_software.lower().startswith("solara"):
    import nbformat

    nb: nbformat.NotebookNode = nbformat.read(settings.domino_notebook_deploy_filename, 4)
    notebook.markdown_cells = {cell["id"]: cell["source"] for cell in nb.cells if cell["cell_type"] == "markdown"}


class AssistantWidget(vy.VuetifyTemplate):
    template_file = (__file__, "assistant.vue")

    menu = traitlets.Any().tag(sync=True, **ipywidgets.widget_serialization)
    cell_id = traitlets.Any().tag(sync=True)
    code = traitlets.Dict(allow_none=True).tag(sync=True)
    code_up = traitlets.Dict(allow_none=True).tag(sync=True)
    notebook_path = traitlets.Unicode(allow_none=True).tag(sync=True)
    save_counter = traitlets.Int(0).tag(sync=True)
    markdown_cells = traitlets.Dict(allow_none=True).tag(sync=True)
    snippet_prefix = traitlets.Unicode(snippets.EDIT_SNIPPET_PREFIX).tag(sync=True)
    snippet_saved_snackbar = traitlets.Bool(False).tag(sync=True)
    snippet_saved_count = traitlets.Int(0).tag(sync=True)
    snippet_edit_mode = traitlets.Bool(False).tag(sync=True)
    snippet_add_dialog_open = traitlets.Bool(False).tag(sync=True)
    snippet_add_header = traitlets.Unicode(allow_none=True).tag(sync=True)
    in_user_install_mode = traitlets.Bool(handle_user_install.in_user_install_mode()).tag(sync=True)
    version = traitlets.Unicode(__version__).tag(sync=True)
    tour = traitlets.Any().tag(sync=True, **ipywidgets.widget_serialization)
    sample_project = traitlets.List().tag(sync=True)

    def __init__(self, *args, **kwargs):
        self._on_saved: Optional[Callable[[bool], None]] = None
        super().__init__(*args, **kwargs)
        notebook.save_notebook = self.save_notebook

    def save_notebook(self, on_saved: Callable[[bool], None]):
        self._on_saved = on_saved
        self.send({"method": "save_notebook", "args": []})
        self.save_counter += 1

    def vue_notebook_saved(self, result):
        self._on_saved and self._on_saved(result)
        self._on_saved = None

    def vue_save_snippet(self, code):
        snippets.save_snippet(code)
        self.snippet_saved_count += 1

    def vue_insert_sample_project(self, _):
        load_data = action.ActionDemo(df_var_out="df_palmer_penguins", df_name="palmerpenguins")
        transform = [
            ActionUseNbLocals(df_var_out="df_palmer_penguins", var_name="df_palmer_penguins"),
            ActionFilter(df_var_out="df_penguins_clean", col="bill_length_mm", dtype="float64", op="!=", value="nan", is_string=False),
            ActionFilter(df_var_out="df_penguins_clean", col="sex", dtype="object", op="!=", value="nan", is_string=False),
        ]
        plot_1 = Viz(
            name="Untitled",
            plot_type="scatter",
            id="910a55b1-de6e-4a22-8387-18e6e92c87c1",
            col_args={"x": "bill_length_mm", "y": "bill_depth_mm", "color": "species", "facet_col": "sex"},
            df_var_name="df_penguins_clean",
            plot_var_name="plot_1",
            crossfilter_enabled=True,
        )
        plot_2 = Viz(
            name="Untitled",
            plot_type="histogram",
            id="c3a304dc-8071-4641-a138-33c903870f26",
            col_args={"x": "island", "y": "bill_length_mm", "histfunc": "avg", "color": None},
            df_var_name="df_penguins_clean",
            plot_var_name="plot_2",
            crossfilter_enabled=True,
        )
        widget_year = WidgetState(
            df_var="df_penguins_clean", widget="Select", column="year", configurable=True, invert=False, mode=">", multiple=True, out_var="widget_year"
        )
        layout = [
            {"i": "plot_1", "w": 12, "h": 15, "x": 0, "y": 9},
            {"i": "df_penguins_clean", "w": 3, "h": 9, "x": 0, "y": 27},
            {"i": "plot_2", "w": 9, "h": 12, "x": 3, "y": 24},
            {"i": "widget_year", "w": 3, "h": 3, "x": 0, "y": 24},
            {"i": "4293a57d", "w": 12, "h": 4, "x": 0, "y": 0},
            {"i": "ec27252e", "w": 6, "h": 5, "x": 0, "y": 4},
            {"i": "f7cd73bb", "w": 6, "h": 5, "x": 6, "y": 4},
        ]

        self.sample_project = [
            {"code": sample_project_md.md1, "type_": "markdown", "id": "4293a57d"},
            {"code": sample_project_md.md2, "type_": "markdown"},
            {"code": load_data.render_code("") + f"\n{load_data.df_var_out}", "meta": dumps(load_data)},
            {"code": sample_project_md.md3, "type_": "markdown"},
            {"code": action.render_code(action.render_code_chunks(transform)), "meta": dumps(actions_store.ActionsState(actions=transform))},
            {"code": sample_project_md.md4, "type_": "markdown"},
            {"code": plot_builder.generate_code(plot_1), "meta": dumps(plot_1)},
            {"code": sample_project_md.md5, "type_": "markdown"},
            {"code": plot_builder.generate_code(plot_2), "meta": dumps(plot_2)},
            {"code": sample_project_md.md6, "type_": "markdown"},
            {"code": _generate_map[widget_year.widget or "mypy"](widget_year), "meta": json.dumps(to_json(widget_year))},
            {"code": sample_project_md.md7, "type_": "markdown"},
            {"code": sample_project_md.md8, "type_": "markdown", "id": "ec27252e"},
            {"code": sample_project_md.md9, "type_": "markdown", "id": "f7cd73bb"},
            {"code": sample_project_md.md10, "type_": "markdown"},
            make_app_code_cell(
                layout,
                {
                    "4293a57d": sample_project_md.md1,
                    "ec27252e": sample_project_md.md8,
                    "f7cd73bb": sample_project_md.md9,
                },
            ),
        ]


class MenuWidget(vy.VuetifyTemplate):
    template_file = (__file__, "menu.vue")

    logo = traitlets.Unicode(util.logo).tag(sync=True)
    items = traitlets.List(default_value=[]).tag(sync=True)
    selected = traitlets.Dict(allow_none=True).tag(sync=True)
    snippet_edit_mode = traitlets.Bool(False).tag(sync=True)
    overlay_open = traitlets.Bool().tag(sync=True)
    disabled = traitlets.Bool(True).tag(sync=True)
    overwrite = traitlets.Bool(False).tag(sync=True)


items = [
    {
        "title": "Load data",
        "icon": "mdi-cloud-upload-outline",
        "action": "load_data",
    },
    {
        "title": "AutoML",
        "icon": "mdi-head-cog-outline",
        "action": "automl",
    },
    {
        "title": "Transformations",
        "icon": "mdi-function-variant",
        "action": "transformations",
    },
    {
        "title": "Visualizations",
        "icon": "mdi-chart-bar-stacked",
        "action": "visualizations",
    },
    {
        "title": "Crossfilter widgets",
        "icon": "mdi-filter",
        "action": "x-widgets",
    },
    {
        "title": "App",
        "icon": "mdi-apps",
        "action": "app",
    },
    {
        "title": "Enter text",
        "icon": "mdi-text",
        "action": "markdown",
    },
    {
        "title": "Insert snippet",
        "icon": "mdi-card-text-outline",
        "action": "snippets",
    },
]


@reacton.component
def Menu(selected, on_selected, decoded, snippet_edit_mode, on_snippet_edit_mode, disabled, overwrite):
    menu_action = None
    if decoded and decoded.__class__ in [action.ActionOpen, action.ActionDownloadDataSource, action.ActionDemo]:
        menu_action = "load_data"
    elif decoded and type(decoded) == MLTask:
        menu_action = "automl"
    elif decoded and isinstance(decoded, actions_store.ActionsState):
        menu_action = "transformations"
    elif decoded and type(decoded) == Viz:
        menu_action = "visualizations"
    elif decoded and isinstance(decoded, crossfilter_widgets.WidgetState):
        menu_action = "x-widgets"
    elif decoded and "type_card_grid_layout" in decoded:
        menu_action = "app"
    elif decoded and "markdown" in decoded:
        menu_action = "markdown"

    overlay_open, set_overlay_open = reacton.use_state(False)

    def close_overlay():
        set_overlay_open(False)

    # note: use_effect is used here over use_memo, so the overlay is closed last
    reacton.use_effect(close_overlay, [selected])

    return MenuWidget.element(
        items=[{"icon": "mdi-pencil", "title": "Edit", "action": menu_action, "edit": True}, *items] if decoded else items,
        selected=selected,
        on_selected=lambda v: v and on_selected(v),
        snippet_edit_mode=snippet_edit_mode,
        on_snippet_edit_mode=on_snippet_edit_mode,
        overlay_open=overlay_open,
        on_overlay_open=set_overlay_open,
        disabled=disabled,
        overwrite=overwrite,
    )


@reacton.component
def MarkdownPanel(edit_action, set_code, open, edit, on_close, overwrite_warning=None):
    def on_apply():
        mixpanel.api.track_with_defaults(
            "inserted code",
            {
                "section": "Enter markdown",
            },
        )
        set_code(
            {
                "code": markdown_text,
                "meta": {},
                "type_": "markdown",
            }
        )
        on_close()

    markdown_text, set_markdown_text = reacton.use_state("")

    def update_markdown():
        if open and edit and edit_action and "markdown" in edit_action:
            set_markdown_text(edit_action["markdown"])
        else:
            set_markdown_text("")

    reacton.use_memo(update_markdown, [edit_action, edit])

    with drawer.RightDrawer(
        open=open,
        on_open=lambda v: on_close() if not v else None,
        title="Markdown",
        edit=bool(edit),
        on_apply=on_apply,
        warning_widget=overwrite_warning,
    ) as main:
        with sol.Div(class_="markdown-drawer", style_="min-width: 800px; max-width: 800px"):
            if open:
                sol.MarkdownEditor(markdown_text, on_value=set_markdown_text)
    return main


class Symbol:
    def __init__(self, name: str):
        self._name = name

    def __repr__(self):
        return self._name


def make_app_code_cell(layout, markdown_cells=None):
    def filter_keys(d: Dict):
        return {k: v for k, v in d.items() if k in ["w", "h", "x", "y"]}

    def layout_item(item: Dict):
        obj: Any = Symbol(item["i"])
        if markdown_cells and item["i"] in markdown_cells:
            obj = Symbol(f'dca.MarkdownFromCell({item["i"]!r})')
        return {"item": obj, **filter_keys(item)}

    layout_with_items = [layout_item(item) for item in layout]
    layout_lines = (",\n" + (" " * 24)).join(str(k) for k in layout_with_items)
    code = textwrap.dedent(
        f"""\
                from domino_code_assist.deploy import Deployer
                import solara


                @solara.component
                def Page():
                    return dca.CardGridLayout([
                        {layout_lines}
                    ])

                Deployer(Page)"""
    )
    return {"code": code, "meta": json.dumps({"type_card_grid_layout": layout})}


@reacton.component
def AppPanel(edit_data, set_code, open, edit, on_close, markdown_cells: Optional[Dict] = None, overwrite_warning=None):
    layout, on_layout = sol.use_state_or_update(open and edit and edit_data and edit_data["type_card_grid_layout"] or [])

    plot_vars = get_vars(lambda v: isinstance(v, plotly.graph_objs._figure.Figure) or isinstance(v, reacton.core.Element))
    dfs = get_vars(lambda v: isinstance(v, (DataFrame, app.ReactiveDf)))

    df_quick_start_created, set_df_quick_start_created = reacton.use_state(False)
    plot_quick_start_created, set_plot_quick_start_created = reacton.use_state(False)

    if not plot_vars and open and "df_quick_start" not in (dfs or []):
        set_df_quick_start_created(True)
        util.nb_locals["df_quick_start"] = data.palmerpenguins()
        dfs = ["df_quick_start"] + (dfs or [])

    if not plot_vars and open and "plot_quick_start" not in plot_vars:
        set_plot_quick_start_created(True)
        import plotly.express as px

        util.nb_locals["plot_quick_start"] = px.scatter(util.nb_locals["df_quick_start"], x="bill_depth_mm", y="bill_length_mm", color="species")
        util.nb_locals["plot_quick_start"].update_layout(margin=dict(l=0, r=0, t=0, b=0))
        plot_vars = ["plot_quick_start"] + plot_vars

    def make_code():
        should_remove_df_quick_start = True
        should_remove_plot_quick_start = True

        qs_used = [True for item in layout if item["i"] == "df_quick_start"]
        if qs_used and df_quick_start_created:
            new_action = action.ActionDemo(df_var_out="df_quick_start", df_name="palmerpenguins")
            set_code({"code": new_action.render_code("df") + f"\n{new_action.df_var_out}", "meta": dumps(new_action), "modifier": "insert-above"})
            should_remove_df_quick_start = False

        qs_plot_used = [True for item in layout if item["i"] == "plot_quick_start"]
        if qs_plot_used and plot_quick_start_created:
            viz = Viz(
                plot_type="scatter",
                col_args={"x": "bill_depth_mm", "y": "bill_length_mm", "color": "species"},
                df_var_name="df_quick_start",
                plot_var_name="plot_quick_start",
                crossfilter_enabled=False,
            )
            set_code({"code": plot_builder.generate_code(viz), "meta": dumps(viz), "modifier": "insert-above"})
            should_remove_plot_quick_start = False

        set_code(make_app_code_cell(layout, markdown_cells))
        mixpanel.api.track_with_defaults("inserted code", {"section": "app"})
        on_close_(should_remove_df_quick_start, should_remove_plot_quick_start)

    def on_close_(should_remove_df_quick_start=True, should_remove_plot_quick_start=True):
        on_close()
        on_layout([])
        if df_quick_start_created and should_remove_df_quick_start:
            util.nb_locals.pop("df_quick_start", None)
        if plot_quick_start_created and should_remove_plot_quick_start:
            util.nb_locals.pop("plot_quick_start", None)

    def on_toggle(on):
        mixpanel.api.track_with_defaults(
            "interaction",
            {
                "section": "app",
                "type": "toggle variable",
                "state": bool(on),
            },
        )

    with drawer.RightDrawer(
        open=open,
        on_open=lambda v: on_close_() if not v else None,
        title="App",
        edit=bool(edit),
        on_apply=make_code,
        width="100%",
        warning_widget=overwrite_warning,
    ) as main:
        if open:
            CardGridLayoutBuilder(obj_vars=plot_vars + (dfs or []), layout=layout, on_layout=on_layout, markdown_cells=markdown_cells, on_toggle=on_toggle).key(
                "grid-layout"
            )

    return main


@reacton.component
def ExceptionGuard(children):
    busy, set_busy = reacton.use_state(False)
    exception, clear_exception = sol.use_exception()
    terminated, set_terminated = reacton.use_state(False)

    def reset():
        set_busy(True)
        clear_exception()
        set_busy(False)

    def terminate():
        set_terminated(True)
        clear_exception()

    if exception:
        with v.Dialog(v_model=bool(exception), on_v_model=lambda v: reset() if not v else None) as dialog:
            with v.Card():
                with v.CardTitle(class_="headline"):
                    v.Html(tag="h3", children=["Unexpected error"])
                with v.CardText():
                    if exception is None or exception.__traceback__ is None:
                        trace = ""
                    else:
                        trace = "".join(reversed(traceback.format_exception(None, exception, exception.__traceback__)[1:]))
                    v.Html(tag="pre", children=[trace], style_="max-height: 300px; overflow: auto;")
                with v.CardActions():
                    sol.Button("Close", color="primary", loading=busy, text=True, on_click=reset)
                    sol.Button("Terminate Code Assistant", color="primary", loading=busy, text=True, on_click=terminate)
        return dialog
    else:
        if terminated:
            sol.Error("Code Assistant terminated. Please restart the notebook. Consider reporting this bug.")
        else:
            return children


@reacton.component
def OverWriteWarning(code, type_):
    with sol.Div(class_="pa-4") as main:
        v.Html(tag="h4", children=[f"Warning: the {'text' if type_ == 'markdown' else 'code'} below will be overwritten."])
        if type_ == "markdown":
            with v.Sheet(elevation=2, class_="pa-2"):
                sol.Markdown(code)
        else:
            CodeHighlightCss()
            Code(
                code_chunks=[code],
                on_event=lambda a: None,
                error=None,
            )
    return main


@reacton.component
def Assistant():
    if settings.domino_code_assist_exception_guard:
        return ExceptionGuard(children=AssistantMain())
    else:
        return AssistantMain()


@reacton.component
def AssistantMain():
    code, set_code = reacton.use_state(cast(Optional[Dict], None))
    code_up, set_code_up = reacton.use_state(cast(Optional[Dict], None))
    selected, set_selected = reacton.use_state(cast(Optional[Dict], None))

    notebook_path, set_notebook_path = reacton.use_state(cast(Optional[str], None))

    def sync_notebook_browser_path():
        if notebook_path:
            domino_code_assist.assistant.notebook.notebook_browser_path.value = notebook_path

    solara.use_memo(sync_notebook_browser_path, dependencies=[notebook_path])

    edit_data, set_edit_data = reacton.use_state(cast(Optional[Dict["str", Any]], None))
    markdown_cells, set_markdown_cells = reacton.use_state(cast(Optional[Dict], None))

    snippet_saved_count, set_snippet_saved_count = reacton.use_state(0)
    snippet_edit_mode, set_snippet_edit_mode = reacton.use_state(False)

    snippet_add_header, set_snippet_add_header = reacton.use_state(cast(Optional[str], None))
    add_dialog_open, set_add_dialog_open = reacton.use_state(False)

    def handle_code_up():
        if code_up and code_up.get("type_") == "markdown":
            set_edit_data({"markdown": code_up["code"]})
        elif code_up and code_up.get("meta"):
            set_edit_data(loads(code_up.get("meta")))
        else:
            set_edit_data(None)

    reacton.use_memo(handle_code_up, [code_up])

    dfs = get_vars(lambda v: isinstance(v, DataFrame))

    def update_markdown():
        notebook.markdown_cells = markdown_cells

    reacton.use_memo(update_markdown, [markdown_cells])

    def on_selected():
        if selected and selected.get("action"):
            mixpanel.api.track_with_defaults(
                "open",
                {
                    "section": selected.get("action"),
                    "edit": bool(selected.get("edit")),
                },
            )

    reacton.use_memo(on_selected, [selected])
    writable_paths = reacton.use_memo(snippets.get_writable_paths, [])

    overwrite = bool(code_up is not None and code_up["code"].strip())

    overwrite_warning = sol.use_memo(
        lambda: OverWriteWarning(code_up["code"], code_up.get("type_")) if code_up and selected and not selected.get("edit") and overwrite else None
    )

    with v.Sheet() as main:
        css.Css()

        if mixpanel.mixpanel_widget:
            sol.Div(children=[mixpanel.mixpanel_widget])
        AssistantWidget.element(
            menu=Menu(
                selected=selected,
                on_selected=set_selected,
                decoded=edit_data,
                snippet_edit_mode=snippet_edit_mode,
                on_snippet_edit_mode=set_snippet_edit_mode,
                disabled=bool(code_up is None or "dca.init()" in code_up["code"]),
                overwrite=overwrite,
            ),
            code=code,
            on_code=set_code,
            on_notebook_path=set_notebook_path,
            on_code_up=set_code_up,
            markdown_cells=markdown_cells,
            on_markdown_cells=set_markdown_cells,
            on_snippet_saved_count=set_snippet_saved_count,
            snippet_edit_mode=snippet_edit_mode,
            snippet_add_header=snippet_add_header,
            snippet_add_dialog_open=add_dialog_open,
            on_snippet_add_dialog_open=set_add_dialog_open,
            tour=TourWidget.element(),
        )

        MarkdownPanel(
            edit_data,
            set_code,
            open=(selected and selected.get("action")) == "markdown",
            edit=selected and selected.get("edit"),
            on_close=lambda: set_selected(None),
            overwrite_warning=overwrite_warning,
        )

        LoadDataDrawer(
            edit_data,
            set_code,
            open=(selected and selected.get("action")) == "load_data",
            edit=selected and selected.get("edit"),
            on_close=lambda: set_selected(None),
            overwrite_warning=overwrite_warning,
        ).meta(ref="load_data")

        AutoMlDrawer(
            edit_data,
            set_code,
            dfs,
            is_open=(selected and selected.get("action")) == "automl",
            edit=selected and selected.get("edit"),
            on_close=lambda: set_selected(None),
            on_load_own_data=lambda: set_selected({"title": "Load data", "icon": "mdi-cloud-upload-outline", "action": "load_data"}),
            overwrite_warning=overwrite_warning,
        ).key("automl-drawer")

        TransformDrawer(
            edit_data,
            set_code,
            dfs,
            open=(selected and selected.get("action")) == "transformations",
            edit=selected and selected.get("edit"),
            on_close=lambda: set_selected(None),
            on_load_own_data=lambda: set_selected({"title": "Load data", "icon": "mdi-cloud-upload-outline", "action": "load_data"}),
            overwrite_warning=overwrite_warning,
        ).key(f'trans-edit-{selected and selected.get("edit")}').meta(ref="transformations")

        VisualizationDrawer(
            cast(Viz, edit_data),
            set_code,
            dfs,
            open=(selected and selected.get("action")) == "visualizations",
            edit=(selected and selected.get("edit")),
            on_close=lambda: set_selected(None),
            on_load_own_data=lambda: set_selected({"title": "Load data", "icon": "mdi-cloud-upload-outline", "action": "load_data"}),
            overwrite_warning=overwrite_warning,
        ).key(f'viz-edit-{selected and selected.get("edit")}').meta(ref="viz")

        crossfilter_widgets.CrossFilterWidgetsPanel(
            cast(Optional[crossfilter_widgets.WidgetState], edit_data),
            set_code,
            dfs,
            open=(selected and selected.get("action")) == "x-widgets",
            edit=selected and selected.get("edit"),
            on_close=lambda: set_selected(None),
            overwrite_warning=overwrite_warning,
        ).key(f'crossfilter-edit-{selected and selected.get("edit")}')

        AppPanel(
            edit_data,
            set_code,
            open=(selected and selected.get("action")) == "app",
            edit=selected and selected.get("edit"),
            on_close=lambda: set_selected(None),
            markdown_cells=markdown_cells,
            overwrite_warning=overwrite_warning,
        )

        SnippetDrawer(
            set_code,
            open=(selected and selected.get("action")) == "snippets",
            on_close=lambda: set_selected(None),
            save_count=snippet_saved_count,
            edit_mode=snippet_edit_mode,
            on_edit_mode=set_snippet_edit_mode,
            writable_paths=writable_paths,
            overwrite_warning=overwrite_warning,
        )

        def on_add_snippet(base_path, path, name):
            set_snippet_add_header(
                f"{snippets.EDIT_SNIPPET_PREFIX} {Path(base_path) / 'snippets' / Path(*[p['id'] for p in path[1:] if not p['leaf']] + [name + '.py'])}\n"
            )
            set_add_dialog_open(False)

        if writable_paths:
            snippets_ui.AddDialog(writable_paths, add_dialog_open, set_add_dialog_open, on_add_snippet)

        def close():
            set_selected(None)

    return main



import reacton.ipyvuetify as v
import solara

from .. import domino_api
from . import object_store, snowflake_redshift
from .big_query import BigQueryPanel
from .state import load_data


@solara.component
def DataSourcePanel(
    columns,
    visible,
):
    load_data.use()
    data_source = load_data.data_source

    datasource_list_result: solara.Result = solara.use_thread(lambda _cancel: domino_api.get_domino_api().get_datasource_list() if visible else None, [visible])

    def get_ds(name):
        if not name:
            return None
        from typing import TYPE_CHECKING

        from domino_code_assist.util import in_dev_mode

        if TYPE_CHECKING:
            from domino.data_sources import DataSourceClient  # pragma: no cover
        else:
            if in_dev_mode():
                from domino_code_assist.domino_lab_mock import DataSourceClient
            else:
                from domino.data_sources import DataSourceClient  # pragma: no cover
        ds_ = DataSourceClient().get_datasource(name)
        data_source.set_type(ds_.datasource_type)
        return ds_

    ds_result: solara.Result = solara.use_thread(lambda: get_ds(data_source.value.name), [data_source.value.name])
    ds = ds_result.value

    def supported(ds_type):
        return ds_type in [*object_store.supported_types, *snowflake_redshift.supported_types, "BigQueryConfig"]

    def text(e):
        return f'{e["name"]} - {e["dataSourceType"].replace("Config", "")}{" - [not yet supported]" if not supported(e["dataSourceType"]) else ""}'

    with solara.Div() as main:
        with v.Row():
            with v.Col():
                if datasource_list_result.error:
                    solara.Error(str(datasource_list_result.error))
                else:
                    items = datasource_list_result.value and [
                        {
                            "text": text(e),
                            "value": e["name"],
                            "disabled": not supported(e["dataSourceType"]),
                        }
                        for e in datasource_list_result.value
                    ]  # type: ignore
                    v.Select(
                        label="Data Source",
                        # TODO: bug in vuetify wrappers
                        items=items,
                        v_model=data_source.value.name,
                        on_v_model=data_source.set_name,
                    ).meta(ref="datasources")
        if data_source.value.name:
            if ds_result.error:
                error = ds_result.error
                error_msg = f"{type(error).__name__}: {str(error)}"
                if isinstance(error, KeyError):
                    error_msg = (
                        error_msg
                        + ". This may be due to an environment with an incompatible version of the Domino-API-client (v"
                        + f"{domino_api.get_domino_api().get_client_version()}) for this Domino version (v{domino_api.get_domino_api().get_domino_version()})."
                    )
                solara.Warning(error_msg)
            elif not ds or ds_result.state == solara.ResultState.RUNNING:
                v.ProgressLinear(indeterminate=True).meta(ref="tables_loading")
            elif ds.datasource_type in object_store.supported_types:
                object_store.Panel(ds)
            elif ds.datasource_type in snowflake_redshift.supported_types:
                snowflake_redshift.Panel(ds)
            elif ds.datasource_type == "BigQueryConfig":
                BigQueryPanel(ds=ds).key("bq" + data_source.value.name)  # type: ignore
            if data_source.value.sample:
                with solara.Div(class_="ml-4"):
                    solara.Warning("This table has too many rows. A sample of this table will be loaded.")

    return main



import os
from typing import List

import requests
from requests.auth import AuthBase, HTTPBasicAuth

from domino_code_assist.domino_api import AppInfo, IDominoApi
from domino_code_assist.settings import settings

DOMINO_TOKEN_FILE_KEY_NAME = "DOMINO_TOKEN_FILE"
DOMINO_USER_API_KEY_KEY_NAME = "DOMINO_USER_API_KEY"
host = settings.domino_alternative_api_host or settings.domino_api_host


class BearerAuth(AuthBase):
    """
    Class for authenticating requests by user supplied token.
    """

    def __init__(self, auth_token=None, domino_token_file=None):
        self.auth_token = auth_token
        self.domino_token_file = domino_token_file

    def _from_token_file(self):
        with open(self.domino_token_file, "r") as token_file:
            return token_file.readline().rstrip()

    def __call__(self, r):
        """
        Override the default __call__ method for the AuthBase base class

        More more info, see:
        https://docs.python-requests.org/en/master/user/advanced/
        """
        auth_token = self._from_token_file() if self.domino_token_file else self.auth_token
        r.headers["Authorization"] = "Bearer " + auth_token
        return r


def get_auth_by_type(api_key=None, auth_token=None, domino_token_file=None):
    """
    Return appropriate authentication object for requests.

    If no authentication credential is provided, the call fails with an AssertError

    Precedence in the case of multiple credentials is:
        1. auth_token string
        2. domino_token_file
        3. api_key
        4. domino_token_file_from_env
        5. api_key_from_env
    """

    if auth_token is not None:
        return BearerAuth(auth_token=auth_token)
    elif domino_token_file is not None:
        return BearerAuth(domino_token_file=domino_token_file)
    elif api_key is not None:
        return HTTPBasicAuth("", api_key)
    else:
        # In the case that no authentication type was passed when this method
        # called, fall back to deriving the auth info from the environment.
        api_key_from_env = os.getenv(DOMINO_USER_API_KEY_KEY_NAME)
        domino_token_file_from_env = os.getenv(DOMINO_TOKEN_FILE_KEY_NAME)
        if api_key_from_env or domino_token_file_from_env:
            return get_auth_by_type(api_key=api_key_from_env, domino_token_file=domino_token_file_from_env)
        else:
            # All attempts failed -- nothing to do but raise an error.
            raise RuntimeError("Unable to authenticate: no authentication provided")


def get(url, **kwargs):
    return requests.Session().get(url, auth=get_auth_by_type(), **kwargs)


def post(url, data, json, **kwargs):
    return requests.Session().post(url, auth=get_auth_by_type(), data=data, json=json, **kwargs)


class Py36DominoApi(IDominoApi):
    def get_datasource_list(self):
        return []

    def get_datasource_names(self):
        return []

    def sync(self, commit_message: str):
        pass

    def get_app_status(self):
        return None

    def app_publish(self, hardwareTierId):
        pass

    def app_unpublish(self):
        pass

    def get_app_id(self):
        pass

    def get_user_info(self):
        return get(f"{host}/v4/users/self").json()

    def get_domino_version(self) -> str:
        return get(f"{host}/version").json().get("version")

    def sync_files(self, commit_message: str):
        return {}

    def get_useable_environments(self):
        return []

    def get_current_environment_id(self):
        return "not-supported"

    def get_app_info(self) -> List[AppInfo]:
        return []



import json
import textwrap
from dataclasses import asdict, dataclass, replace
from typing import List, Optional

import reacton.ipyvuetify as v
import solara
from solara.lab.utils.dataframe import df_py_types

from domino_code_assist import util
from domino_code_assist.assistant import drawer, mixpanel


@dataclass(frozen=True)
class WidgetState:
    df_var: Optional[str] = None
    widget: Optional[str] = None
    column: Optional[str] = None
    configurable: bool = True
    invert: bool = False
    mode: str = "=="
    multiple: bool = False
    out_var: Optional[str] = None


@solara.component
def CrossFilterWidgetsPanel(edit_data: Optional[WidgetState], set_code, dfs, open, edit, on_close, overwrite_warning=None):
    state, set_state = solara.use_state_or_update(edit_data if open and edit and edit_data else WidgetState(out_var=util.generate_var_name("widget")))

    def set_plot_var(val):
        set_state(lambda s: replace(s, out_var=val))

    available_widgets = list(_generate_map.keys())

    columns = solara.use_memo(lambda: util.nb_locals[state.df_var].columns.tolist() if state.df_var else [], [state.df_var])
    types = solara.use_memo(lambda: [i.name for i in util.nb_locals[state.df_var].dtypes.tolist()] if state.df_var else [], [state.df_var])
    column_items = [{"text": f"{c} ({t})", "value": c} for c, t in zip(columns, types)]

    def on_apply():
        assert state.widget is not None
        mixpanel.api.track_with_defaults(
            "inserted code",
            {
                "section": "Cross-filter widgets",
            },
        )
        set_code({"code": _generate_map[state.widget](state), "meta": json.dumps(to_json(state))})
        on_close()

    valid_conf = True
    invalid_reason = ""
    if state.widget == "Slider":
        if state.df_var and state.column:
            df = util.nb_locals[state.df_var]
            py_type = df_py_types(df)
            if py_type[state.column] not in (int, float):
                valid_conf = False
                invalid_reason = "Slider widget only supports numerical columns"

    with drawer.RightDrawer(
        open=open,
        on_open=lambda v: on_close() if not v else None,
        title="Crossfilter Widgets",
        edit=bool(edit),
        on_apply=on_apply,
        apply_disabled=not ((state.df_var and (state.widget == "Report" or state.column) and state.widget and state.out_var) and valid_conf),
        show_var_out=True,
        var_out=state.out_var or "",
        on_var_out=set_plot_var,
        width="768px",
        warning_widget=overwrite_warning,
    ) as main:
        with solara.Div():
            with solara.Div(class_="dca-transform-header", style_="margin-left: -24px; margin-right: -24px; margin-top: -4px; padding: 8px 80px;"):
                v.Select(label="Dataframe", items=dfs, v_model=state.df_var, on_v_model=lambda v: set_state(lambda s: replace(s, df_var=v)))
            if state.df_var:
                v.Select(label="Widget", items=available_widgets, v_model=state.widget, on_v_model=lambda v: set_state(replace(state, widget=v)))
                if state.widget:
                    if state.widget != "Report":
                        v.Select(label="Column", items=column_items, v_model=state.column, on_v_model=lambda v: set_state(replace(state, column=v)))
                        v.Switch(label="Invert filter", v_model=state.invert, on_v_model=lambda v: set_state(replace(state, invert=v)))
                        v.Switch(label="Configurable", v_model=state.configurable, on_v_model=lambda v: set_state(replace(state, configurable=v)))
                        if state.widget == "Slider":
                            solara.Div(children=["Mode"], style_="color: rgba(0, 0, 0, 0.6)")
                            with solara.ToggleButtonsSingle(value=state.mode, on_value=lambda v: set_state(replace(state, mode=v))):  # type: ignore
                                solara.Button(icon_name="mdi-code-equal", icon=True, value="==")
                                solara.Button(icon_name="mdi-code-not-equal", icon=True, value="!=")
                                solara.Button(icon_name="mdi-code-less-than", icon=True, value="<")
                                solara.Button(icon_name="mdi-code-less-than-or-equal", icon=True, value="<=")
                                solara.Button(icon_name="mdi-code-greater-than", icon=True, value=">")
                                solara.Button(icon_name="mdi-code-greater-than-or-equal", icon=True, value=">=")
                        if state.widget == "Select":
                            v.Switch(label="Multiple", v_model=state.multiple, on_v_model=lambda v: set_state(replace(state, multiple=v)))
                if not valid_conf:
                    solara.Error(invalid_reason, classes=["my-4"])
    return main


_default = WidgetState()


def _param_list_str(state: WidgetState, names: List[str]):
    def param_str(name: str):
        value = getattr(state, name)
        quoted = f'"{value}"' if isinstance(value, str) else value

        return f", {name}={quoted}" if value != getattr(_default, name) else ""

    return "".join([param_str(name) for name in names])


def _generate_select_widget(state: WidgetState):
    return textwrap.dedent(
        f"""\
        from solara import CrossFilterSelect

        {state.out_var} = CrossFilterSelect({state.df_var}{_param_list_str(state, ["column", "configurable", "invert", "multiple"])})
        {state.out_var}
        """
    )


def _generate_slider_widget(state: WidgetState):
    return textwrap.dedent(
        f"""\
        from solara import CrossFilterSlider

        {state.out_var} = CrossFilterSlider({state.df_var}{_param_list_str(state, ["column", "configurable", "invert", "mode"])})
        {state.out_var}
        """
    )


def _generate_report_widget(state: WidgetState):
    return textwrap.dedent(
        f"""\
        from solara import CrossFilterReport

        {state.out_var} = CrossFilterReport({state.df_var})
        {state.out_var}
        """
    )


_generate_map = {
    "Select": _generate_select_widget,
    "Slider": _generate_slider_widget,
    "Report": _generate_report_widget,
}


def to_json(obj):
    if isinstance(obj, WidgetState):
        return {"type_widget_state": asdict(obj)}


def from_json(dct):
    if "type_widget_state" in dct:
        dct = dct["type_widget_state"]
        return WidgetState(**dct)



import math
from typing import Optional, Tuple, Union, cast

import reacton
import reacton.ipyvuetify as v
import solara
from solara.components.code_highlight_css import CodeHighlightCss

import domino_code_assist.action as action
import domino_code_assist.actions_store as actions_store
from domino_code_assist import action_ui
from domino_code_assist.code import Code
from domino_code_assist.components import DataFrame
from domino_code_assist.transform import transform_action_ui
from domino_code_assist.util import DominoHeader


@reacton.component
def TransformPanel(state, dispatch, on_reset, on_save, breakpoints, assistant_mode=False):
    show_code, set_show_code = reacton.use_state(False)
    show_table, set_show_table = reacton.use_state(True)
    filter_like, set_filter_like = reacton.use_state(cast(Optional[Tuple[str, int]], None))

    def _get_for_index(colection, index):
        if index is None:
            return colection[-1]
        else:
            return colection[index] if 0 <= index < len(colection) else []

    def get_column_names(index):
        return _get_for_index(state.columns_for_action, index)

    def get_dtypes(index):
        return _get_for_index(state.dtypes_for_action, index)

    def reset():
        dispatch((actions_store.e_reset, None))
        on_reset()

    def on_code_view_event(event):
        if not event:
            return
        index = event["index"]
        if isinstance(state.actions[0], action.ActionUseNbLocals):
            index += 1
        name = event["name"]
        if name == "delete":
            dispatch((actions_store.e_delete_index, index))
        if name == "edit":
            dispatch((actions_store.e_edit_index, index))
        if name == "insert":
            dispatch((actions_store.e_insert_index, index))

    add_state, set_add_state = reacton.use_state(
        cast(Union[action.ActionFilter, action.ActionSelectColumns, action.ActionDropColumns, action.ActionGroupBy, action.ActionMerge], action.ActionFilter())
    )
    dialog_state, set_dialog_state = reacton.use_state(cast(action.Action, action.ActionFilter()))

    def on_edit_index():
        if state.index_of_edit_action is not None:
            set_dialog_state(state.actions[state.index_of_edit_action])
        if state.index_of_insert_action is not None:
            set_dialog_state(action.ActionFilter())

    reacton.use_memo(on_edit_index, [state.index_of_edit_action, state.index_of_insert_action])

    with v.Sheet() as main:
        if state.index_of_edit_action is not None:
            index = state.index_of_edit_action
            with v.Dialog(v_model=True, on_v_model=lambda open: dispatch((actions_store.e_edit_index, None)) if not open else None, width="50vw"):
                with v.Card().meta(ref="EditAction"):
                    DominoHeader(title="Edit transformation")
                    with v.CardText():
                        action_ui.EditAction(
                            columns=get_column_names(index - 1),
                            dtypes=get_dtypes(index - 1),
                            on_action=set_dialog_state,
                            action_obj=dialog_state,
                        ).meta(ref="EditAction")
                    with v.CardActions():
                        v.Spacer()
                        solara.Button("edit transformation", color="primary", on_click=lambda: dispatch((actions_store.e_edit_action, dialog_state)))
        if state.index_of_insert_action is not None:
            with v.Dialog(v_model=True, on_v_model=lambda open: dispatch((actions_store.e_insert_index, None)) if not open else None, width="50vw"):
                with v.Card().meta(ref="insert_panel"):
                    DominoHeader(title="Insert transformation")
                    with v.CardText():
                        transform_action_ui.TransformActionSelectPanel(
                            columns=get_column_names(state.index_of_insert_action),
                            dtypes=get_dtypes(state.index_of_insert_action),
                            action_=dialog_state,
                            on_action=set_dialog_state,
                            on_action_type=lambda t: set_dialog_state(t()),
                        )
                    with v.CardActions():
                        v.Spacer()
                        solara.Button("insert transformation", color="primary", on_click=lambda: dispatch((actions_store.e_insert_action, dialog_state)))
        if filter_like is not None:
            with v.Dialog(v_model=True, on_v_model=lambda open: set_filter_like(None) if not open else None, width="50vw"):
                with v.Card():
                    DominoHeader(title="Filter like")
                    with v.CardText():
                        column, index = filter_like
                        value = state.df_wrapper.df.iloc[index][column]
                        if math.isnan(value):
                            value = "NaN"
                        value = str(value)
                        action_val = action.ActionFilter(col=column, dtype=get_dtypes(None)[column], value=value, op="==")
                        transform_action_ui.FilterPanel(
                            columns=get_column_names(None),
                            dtypes=get_dtypes(None),
                            on_action=lambda action: [set_filter_like(None), dispatch((actions_store.e_insert_action, action))],
                            fill=action_val,
                        ).key(f"filter_like {action_val.__hash__()}")
        with v.Row(no_gutters=not breakpoints):
            with v.Col(sm=12, lg=5 if breakpoints else 12):
                with solara.Div():
                    if show_table and state.df_wrapper.has_df:
                        DataFrame(
                            state.df_wrapper.df,
                            on_drop_column=lambda column: dispatch((actions_store.e_insert_action, action.ActionDropColumns(columns=[column]))),
                            on_filter_value=lambda col, row: set_filter_like((col, row)),
                        ).meta(ref="DataTable").key("DataTable")

                    with solara.Div():

                        def on_action(action_):
                            dispatch((actions_store.e_insert_action, (idx, action_)))
                            set_add_state(action.ActionFilter())

                        idx = len(state.actions) - 1

                        if not state.error:
                            with solara.Div(style_="max-width: 600px; margin-left: auto; margin-right: auto;").meta(ref="insert_panel"):
                                transform_action_ui.TransformActionSelectPanel(
                                    columns=get_column_names(idx),
                                    dtypes=get_dtypes(idx),
                                    action_=add_state,
                                    on_action=set_add_state,
                                    on_action_type=lambda t: set_add_state(t()),
                                )
                                with solara.Div(style_="display: flex"):
                                    v.Spacer()
                                    solara.Button(
                                        outlined=True,
                                        color="primary",
                                        on_click=lambda: on_action(add_state),
                                        disabled=not add_state.is_valid(),
                                        children=["add transformation"],
                                    )
                    with v.CardActions(class_="ma-0"):
                        if not assistant_mode:
                            solara.Button(
                                icon_name="mdi-delete",
                                on_click=lambda: reset(),
                                icon=True,
                            )
                            solara.Button(
                                icon=True,
                                icon_name="mdi-content-copy",
                                disabled=True,
                            )
                            solara.Button(
                                icon_name="mdi-content-save",
                                on_click=on_save,
                                icon=True,
                            )
                        solara.Button(
                            icon_name="mdi-undo",
                            on_click=lambda: dispatch((actions_store.e_undo, None)),
                            icon=True,
                            disabled=not state.undo_stack,
                        ).meta(ref="btn_undo")
                        solara.Button(
                            icon_name="mdi-redo",
                            on_click=lambda: dispatch((actions_store.e_redo, None)),
                            icon=True,
                            disabled=len(state.redo_stack) == 0,
                        ).meta(ref="btn_redo")
                        v.Switch(label="Show table", v_model=show_table, on_v_model=set_show_table, class_="mx-4")
                        v.Switch(label="Show code", v_model=show_code, on_v_model=set_show_code, class_="mx-4")

                    if state.error:
                        v.Alert(
                            type="error",
                            text=True,
                            prominent=True,
                            icon="mdi-alert",
                            children=[f"{state.error[1]}"],
                        )
                    if show_code or state.error:
                        CodeHighlightCss()
                        Code(code_chunks=state.code_chunks, on_event=on_code_view_event, error=state.error, assistant_mode=assistant_mode)
    return main



from playwright.sync_api import Page


class LoadDataHelper:
    def __init__(self, page: Page) -> None:
        self.page = page

    @property
    def menu_item(self):
        return self.page.locator('div[role="listbox"] >> text=Load data')

    @property
    def dialog(self):
        return self.page.locator(".v-dialog__content--active")

    @property
    def project_files(self):
        return ProjectFileHelper(self.page, self)

    @property
    def insert_code(self):
        return self.dialog.locator('button:has-text("Run")')


class ProjectFileHelper:
    def __init__(self, page: Page, load_data: LoadDataHelper) -> None:
        self.page = page
        self.load_data = load_data

    @property
    def tab_item(self):
        return self.load_data.dialog.locator("text=Project Files")

    @property
    def file_browser(self):
        return self.load_data.dialog.locator(".solara-file-browser")



from dataclasses import asdict, dataclass
from typing import Dict, List, Optional

import solara
from solara.lab import Ref

from domino_code_assist import util


@dataclass(frozen=True)
class TaskInfo:
    name: str
    arg: str
    description: str
    image: str = "placeholder.png"
    soon: bool = False


AVAILABLE_TASKS = [
    TaskInfo(
        name="Classification",
        arg="classification",
        description="Tag and categorize any kind of data so that it can be better understood and analyzed.",
        image="icon_classification.svg",
    ),
    TaskInfo(
        name="Regression",
        arg="regression",
        description="Estimate the relationships between variables and output a numeric value.",
        image="icon_regression.svg",
    ),
    TaskInfo(
        name="Forecasting",
        arg="ts_forecast",
        description="Predict future conditions by considering historical data.",
        image="icon_forecasting.svg",
    ),
    TaskInfo(
        name="Forecasting (classification)",
        arg="ts_forecast_classification",
        description="Predict future conditions by considering historical data.",
        image="icon_forecasting.svg",
    ),
]


def get_task_arg_for_name(name):
    for task in AVAILABLE_TASKS:
        if task.name == name:
            return task.arg
    return None


@dataclass(frozen=True)
class TaskConfig:
    df_var_name: Optional[str] = None
    experiment_name: Optional[str] = None
    exclude_features: Optional[List[str]] = None
    target: Optional[str] = None
    optimization_metric: str = "accuracy"
    max_iterations: Optional[int] = None
    max_time: Optional[int] = None
    time_unit: str = "Minutes"
    period: Optional[int] = None
    auto: bool = True
    custom_learners: Optional[Dict] = None
    custom_learners_enabled: Optional[List[str]] = None
    time_column: Optional[str] = None
    resampling_strategy: str = "Auto"
    split_ratio: str = "0.1"
    folds: str = "5"


class ReactiveTaskConfig(solara.lab.Reactive[TaskConfig]):
    def is_valid(self):
        return bool(
            self.value.df_var_name
            and self.value.target
            and self.value.experiment_name
            and (
                self.value.resampling_strategy == "Auto"
                or (self.value.resampling_strategy == "Holdout" and util.to_float(self.value.split_ratio) is not None)
                or (self.value.resampling_strategy == "Cross validation" and util.to_int(self.value.folds) is not None)
            )
        )

    def get_features(self, all_columns):
        return [c for c in all_columns if c not in (self.value.exclude_features or []) and c != self.value.target] if all_columns is not None else None

    def set_target(self, target):
        if target in (self.value.exclude_features or []):
            Ref(self.fields.exclude_features).set([f for f in (self.value.exclude_features or []) if f != target])
        Ref(self.fields.target).set(target)


task_config_store = ReactiveTaskConfig(TaskConfig())


@dataclass(frozen=True)
class MLTask:
    task_type: Optional[str] = None
    task_config: Optional[TaskConfig] = None


class ReactiveMLTask(solara.lab.Reactive[MLTask]):
    pass


ml_task = ReactiveMLTask(MLTask())


def to_json(obj: MLTask):
    if isinstance(obj, MLTask):
        return {
            "type_ml": {
                "task_type": obj.task_type,
                "task_config": {"type": obj.task_config.__class__.__name__, "data": asdict(obj.task_config)} if obj.task_config else None,
            }
        }


def from_json(dct):
    if "type_ml" in dct:
        task_type = dct["type_ml"]["task_type"]
        task_config = dct["type_ml"]["task_config"]
        if task_config:
            task_config = globals()[task_config["type"]](**task_config["data"])
        res = MLTask(task_type=task_type, task_config=task_config)

        return res



# isort: skip_file
# type: ignore
import domino_code_assist as dca
from domino_code_assist.data import palmerpenguins


def display(x):
    pass


___df_var_name___ = palmerpenguins()
___exclude_features___ = []
___time_budget___ = 600
___max_iterations___ = 50
___period___ = 2
___estimator_list___ = []
___custom_hp___ = {}
# ___TEMPLATE_START___ #
from flaml import AutoML
import mlflow.sklearn
from mlflow.models.signature import infer_signature
import pandas as pd

experiment_name = "___experiment_name___"

label = "___label___"
exclude_features = ___exclude_features___

df_automl = ___df_var_name___.drop(exclude_features, axis=1).copy()

display(df_automl)

experiment = mlflow.set_experiment(experiment_name)

estimator_list = ___estimator_list___
custom_hp = ___custom_hp___

# No runs can be active before we start. We will create a run
while mlflow.active_run():
    mlflow.end_run()

mlflow.autolog(disable=True)

with mlflow.start_run() as run:
    run_name = run.data.tags["mlflow.runName"]
    print("Run name: ", run_name)
    display(dca.automl.OpenExperiment(experiment_name))
    automl = AutoML(metric="___optimization_metric___")
    automl.fit(dataframe=df_automl, label=label, task="___ml_task___", period=___period___, time_budget=___time_budget___, max_iter=___max_iterations___, custom_hp=custom_hp, estimator_list=estimator_list, eval_method=None)
    # Log the best model
    train = df_automl.drop(label, axis=1)[:5]
    predictions = pd.DataFrame({label: automl.predict(train)})
    mlflow.sklearn.log_model(automl, "model", signature=infer_signature(train, predictions), input_example=train)
    # Find the best run
    best_run = mlflow.search_runs(
        experiment_ids=[experiment.experiment_id],
        filter_string=f'params.best_config = "{automl.best_config}"',
    )
    if len(best_run.run_id):
        best_run_id = best_run.run_id[0]
        mlflow.set_tag("Best run", best_run_id)

        # Copy the params and metrics from the winning run
        run_params = best_run[[c for c in best_run.columns if c.startswith("params.")]].to_dict()
        run_params = {k.lstrip("params."): v[0] for k, v in run_params.items()}
        mlflow.log_params(run_params)

        run_metrics = best_run[[c for c in best_run.columns if c.startswith("metrics.")]].to_dict()
        run_metrics = {k.lstrip("metrics."): v[0] for k, v in run_metrics.items()}
        mlflow.log_metrics(run_metrics)
    else:
        print("No best run found.")

    # Log the notebook cell execution history for reproducibility
    dca.mlflow_log_notebook(run_name)

# Set the wining run to be tagged as the winner
# And log the model
if len(best_run.run_id):
    with mlflow.start_run(best_run_id):
        mlflow.set_tag("Status", "Winner")
        mlflow.set_tag("Winner", True)
        mlflow.sklearn.log_model(automl, "model", signature=infer_signature(train, predictions), input_example=train)



from typing import Callable, TypeVar, Union

T = TypeVar("T")
U = TypeVar("U")


def use_reducer_addon(reduce: Callable[[T, U], T], set_state: Callable[[Union[T, Callable[[T], T]]], None]) -> Callable[[U], None]:
    def dispatch(action: U):
        def state_updater(state: T):
            return reduce(state, action)

        set_state(state_updater)

    return dispatch



from typing import Optional, cast

import pandas as pd
import reacton
import solara
from solara.lab import Ref

import domino_code_assist
from domino_code_assist import action, data, util
from domino_code_assist.assistant import drawer, mixpanel
from domino_code_assist.util import generate_var_name

from ..df_select import DfSelect
from ..serialize import dumps
from . import plot_builder
from .state import EqWrapper, Viz, viz


@solara.component
def VisualizationDrawer(edit_data: Optional[Viz], set_code, dfs, open, edit, on_close, on_load_own_data, overwrite_warning=None):
    viz.use()
    df_ref = solara.use_ref(cast(pd.DataFrame, None))
    quick_start_created, set_quick_start_created = reacton.use_state(False)

    def set_df_var_name(value):
        if value == "Load more data":
            close()
            on_load_own_data()
            return
        if value == "df_quick_start" and value not in dfs:
            # pre populate notebook scope
            set_quick_start_created(True)
            util.nb_locals["df_quick_start"] = data.palmerpenguins()

        mixpanel.api.track_with_defaults(
            "interaction",
            {
                "section": "visualizations",
                "type": "selected dataframe",
            },
        )
        Ref(viz.fields.df_var_name).value = value

    def set_visualization():
        if viz.value.df_var_name == "df_quick_start" and quick_start_created:
            set_quick_start_created(False)
            new_action = action.ActionDemo(df_var_out="df_quick_start", df_name="palmerpenguins")
            set_code({"code": new_action.render_code("df") + f"\n{new_action.df_var_out}", "meta": dumps(new_action), "modifier": "insert-above"})
            set_quick_start_created(False)

        code_str = plot_builder.generate_code(viz.value)
        set_code({"code": code_str, "meta": dumps(viz.value)})
        mixpanel.api.track_with_defaults("inserted code", {"section": "visualizations"})
        close()

    def close():
        on_close()
        if quick_start_created:
            set_quick_start_created(False)
            util.nb_locals.pop("df_quick_start", None)

    def init_plot_state():
        if open and edit_data and edit:
            viz.set(edit_data)
        else:
            viz.set(Viz(plot_var_name=generate_var_name("var")))

    solara.use_memo(init_plot_state, [open])

    def handle_var_name():
        df_ref.current = (
            EqWrapper(domino_code_assist.util.nb_locals[viz.value.df_var_name]) if viz.value.df_var_name and viz.value.df_var_name != "Load more data" else None
        )

    solara.use_memo(handle_var_name, [viz.value.df_var_name])
    plot_var_name = viz.value.plot_var_name

    extended_dfs = dfs
    if len(dfs) == 0:
        extended_dfs = ["df_quick_start"] + extended_dfs
    if not edit:
        extended_dfs = extended_dfs + [{"divider": True}, "Load more data"]

    def pre_select_df():
        if open and viz.value.df_var_name is None:
            set_df_var_name(dfs[0] if dfs else "df_quick_start")

    reacton.use_memo(pre_select_df, [open])

    def pre_select_plot_vars():
        if open and viz.value.df_var_name == "df_quick_start" and viz.value.plot_type is None:
            Ref(viz.fields.plot_type).set("scatter")
            Ref(viz.fields.col_args).set(
                {
                    "x": "bill_depth_mm",
                    "y": "bill_length_mm",
                    "color": "species",
                }
            )

    reacton.use_memo(pre_select_plot_vars, [viz.value.df_var_name, open])

    with drawer.RightDrawer(
        open=open,
        on_open=lambda v: close() if not v else None,
        title="Visualization",
        edit=bool(edit),
        on_apply=set_visualization,
        apply_disabled=not viz.value.df_var_name,
        show_var_out=True,
        var_out=plot_var_name or "",
        on_var_out=Ref(viz.fields.plot_var_name).set,
        width="1024px",
        warning_widget=overwrite_warning,
    ) as main:
        DfSelect(value=viz.value.df_var_name, label="DataFrame", on_value=set_df_var_name, items=extended_dfs or [])
        if viz.value.df_var_name:
            plot_builder.PlotBuilder(
                df_ref.current,
                df_ref.current and df_ref.current.get().columns.tolist(),
                min_height=None,
            )
    return main



optimization_metric_options = [
    ["auto", "Automatic optimization metric."],
    ["accuracy", "Accuracy as the corresponding metric to minimize."],
    ["log_loss", "Default metric for multiclass classification."],
    ["r2", "R2_score as the corresponding metric to minimize. Default metric for regression."],
    ["rmse", "Root mean squared error."],
    ["mse", "Mean squared error."],
    ["mae", "mean absolute error."],
    ["mape", "Mean absolute percentage error."],
    ["roc_auc", "Roc_auc_score. Default metric for binary classification."],
    ["roc_auc_ovr", """Roc_auc_score with multi_class="ovr"."""],
    ["roc_auc_ovo", """Roc_auc_score with multi_class="ovo"."""],
    ["roc_auc_weighted", """Roc_auc_score with average="weighted"."""],
    ["roc_auc_ovr_weighted", """Roc_auc_score with multi_class="ovr" and average="weighted"."""],
    ["roc_auc_ovo_weighted", """Roc_auc_score with multi_class="ovo" and average="weighted"."""],
    ["f1", "F1_score."],
    ["micro_f1", """F1_score with average="micro"."""],
    ["macro_f1", """F1_score with average="macro"."""],
    ["ap", "Average_precision_score."],
    ["ndcg", "Ndcg_score."],
    ["ndcg@k", "Ndcg_score@k. k is an integer."],
]

common_optimization_metric_options = [
    [k, v]
    for k, v in optimization_metric_options
    if k in ["auto", "accuracy", "roc_auc", "roc_auc_weighted", "f1", "micro_f1", "macro_f1", "log_loss", "mae", "r2"]
]

# See: https://microsoft.github.io/FLAML/docs/Use-Cases/Task-Oriented-AutoML#estimator-and-search-space
hyperparameters = [
    {
        "id": "lgbm",
        "name": "LGBMEstimator",
        "tasks": ["classification", "regression", "rank", "ts_forecast", "ts_forecast_classification"],
        "hyper-parameters": {
            "n_estimators": {"type": "int"},
            "num_leaves": {"type": "int"},
            "min_child_samples": {"type": "int"},
            "learning_rate": {"type": "float"},
            "log_max_bin": {"type": "int"},
            "colsample_bytree": {"type": "float"},
            "reg_alpha": {"type": "float"},
            "reg_lambda": {"type": "float"},
        },
        "hyper-meta": {"log_max_bin": {"comment": "(logarithm of (max_bin + 1) with base 2)"}},
    },
    {
        "id": "xgboost",
        "name": "XGBoostSkLearnEstimator",
        "tasks": ["classification", "regression", "rank", "ts_forecast", "ts_forecast_classification"],
        "hyper-parameters": {
            "n_estimators": {"type": "int"},
            "max_leaves": {"type": "int"},
            "min_child_weight": {"type": "float"},
            "learning_rate": {"type": "float"},
            "subsample": {"type": "float"},
            "colsample_bylevel": {"type": "float"},
            "colsample_bytree": {"type": "float"},
            "reg_alpha": {"type": "float"},
            "reg_lambda": {"type": "float"},
        },
    },
    {
        "id": "xgb_limitdepth",
        "name": "XGBoostLimitDepthEstimator",
        "tasks": ["classification", "regression", "rank", "ts_forecast", "ts_forecast_classification"],
        "hyper-parameters": {
            "n_estimators": {"type": "int"},
            "max_depth": {"type": "int"},
            "min_child_weight": {"type": "float"},
            "learning_rate": {"type": "float"},
            "subsample": {"type": "float"},
            "colsample_bylevel": {"type": "float"},
            "colsample_bytree": {"type": "float"},
            "reg_alpha": {"type": "float"},
            "reg_lambda": {"type": "float"},
        },
    },
    {
        "id": "rf",
        "name": "RandomForestEstimator",
        "tasks": ["classification", "regression", "ts_forecast", "ts_forecast_classification"],
        "hyper-parameters": {
            "n_estimators": {"type": "int"},
            "max_features": {"type": ["sqrt", "log2", "None"]},
            "max_leaves": {"type": "int"},
            "criterion": {"type": ["gini" "entropy" "log_loss"]},
        },
        "hyper-meta": {"criterion": {"tasks": ["classification"]}},
        "extra": "Starting from v1.1.0, it uses a fixed random_state by default.",
    },
    {
        "id": "extra_tree",
        "name": "ExtraTreesEstimator",
        "tasks": ["classification", "regression", "ts_forecast", "ts_forecast_classification"],
        "hyper-parameters": {
            "n_estimators": {"type": "int"},
            "max_features": {"type": ["sqrt", "log2", "None"]},
            "max_leaves": {"type": "int"},
            "criterion": {"type": ["gini" "entropy" "log_loss"]},
        },
        "hyper-meta": {"criterion": {"tasks": ["classification"]}},
        "extra": "Starting from v1.1.0, it uses a fixed random_state by default.",
    },
    {
        "id": "lrl1",
        "readable_name": "logistic regression L1",
        "name": "LRL1Classifier (sklearn.LogisticRegression with L1 regularization)",
        "tasks": ["classification"],
        "hyper-parameters": {
            "C": {"type": "float"},
        },
    },
    {
        "id": "lrl2",
        "readable_name": "logistic regression L2",
        "name": "LRL2Classifier (sklearn.LogisticRegression with L2 regularization)",
        "tasks": ["classification"],
        "hyper-parameters": {
            "C": {"type": "float"},
        },
    },
    {
        "id": "catboost",
        "name": "CatBoostEstimator",
        "tasks": ["classification", "regression"],
        "hyper-parameters": {
            "early_stopping_rounds": {"type": "int"},
            "learning_rate": {"type": "float"},
            "n_estimators": {"type": "int"},
        },
    },
    {
        "id": "kneighbor",
        "name": "KNeighborsEstimator",
        "tasks": ["classification", "regression"],
        "hyper-parameters": {
            "n_neighbors": {"type": "int"},
        },
    },
    {
        "id": "prophet",
        "name": "Prophet",
        "tasks": ["ts_forecast"],
        "hyper-parameters": {
            "changepoint_prior_scale": {"type": "float"},
            "seasonality_prior_scale": {"type": "float"},
            "holidays_prior_scale": {"type": "float"},
            "seasonality_mode": {"type": ["additive", "multiplicative"]},
        },
    },
    {
        "id": "arima",
        "name": "ARIMA",
        "tasks": ["ts_forecast"],
        "hyper-parameters": {
            "p": {"type": "int"},
            "d": {"type": "int"},
            "q": {"type": "int"},
        },
    },
    {
        "id": "sarimax",
        "name": "SARIMAX",
        "tasks": ["ts_forecast"],
        "hyper-parameters": {
            "p": {"type": "int"},
            "d": {"type": "int"},
            "q": {"type": "int"},
            "P": {"type": "int"},
            "D": {"type": "int"},
            "Q": {"type": "int"},
            "s": {"type": "int"},
        },
    },
    {
        "id": "transformer",
        "name": "Huggingface transformer models",
        "tasks": ["seq-classification", "seq-regression", "multichoice-classification", "token-classification", "summarization"],
        "hyper-parameters": {
            "learning_rate": {"type": "float"},
            "num_train_epochs": {"type": "int"},
            "per_device_train_batch_size": {"type": "int"},
            "warmup_ratio": {"type": "float"},
            "weight_decay": {"type": "float"},
            "adam_epsilon": {"type": "float"},
            "seed": {"type": "int"},
        },
    },
    {
        "id": "temporal_fusion_transformer",
        "name": "TemporalFusionTransformerEstimator",
        "tasks": ["ts_forecast_panel"],
        "hyper-parameters": {
            "gradient_clip_val": {"type": "float"},
            "hidden_size": {"type": "int"},
            "hidden_continuous_size": {"type": "int"},
            "attention_head_size": {"type": "int"},
            "dropout": {"type": "float"},
            "learning_rate": {"type": "float"},
        },
        "extra": "There is a known issue with pytorch-forecast logging.",
    },
]


def get_estimator_ids_for_task(task):
    return [item["id"] for item in hyperparameters if task in item["tasks"]]


def get_hyper_parameters_for_task(estimator, task):
    def valid(param):
        for_tasks = estimator.get("hyper-meta", {}).get(param, {}).get("tasks")
        return for_tasks is None or task in for_tasks

    return [param for param in estimator["hyper-parameters"].keys() if valid(param)]



from playwright.sync_api import Page


class AppHelper:
    def __init__(self, page: Page) -> None:
        self.page = page

    @property
    def menu_item(self):
        return self.page.locator('div[role="listbox"] >> text=App')

    @property
    def dialog(self):
        return self.page.locator(".v-dialog__content--active")

    @property
    def insert_code(self):
        return self.dialog.locator('button:has-text("Run")')

    @property
    def preview(self):
        return self.page.locator('button:has-text("Preview")')

    def insert_object(self, name):
        self.page.locator(f'_vue=v-switch[label="{name}"]').click()



from dataclasses import dataclass
from typing import List, Optional

import humanize
import reacton
import reacton.ipyvuetify as v
import solara as sol
from solara.components.sql_code import SqlCode
from solara.lab import Ref

from .state import ReactiveDataSource, load_data


@dataclass(frozen=True)
class SchemasResult:
    default_region: Optional[str]
    schemas: List[str]


regions = [
    "default",
    "us",
    "eu",
    "us-central1",
    "us-west4",
    "us-west2",
    "northamerica-northeast1",
    "us-east4",
    "us-west1",
    "us-west3",
    "southamerica-east1",
    "southamerica-west1",
    "us-east1",
    "northamerica-northeast2",
    "europe-west1",
    "europe-north1",
    "europe-west3",
    "europe-west2",
    "europe-southwest1",
    "europe-west8",
    "europe-west4",
    "europe-west9",
    "europe-central2",
    "europe-west6",
    "asia-south2",
    "asia-east2",
    "asia-southeast2",
    "australia-southeast2",
    "asia-south1",
    "asia-northeast2",
    "asia-northeast3",
    "asia-southeast1",
    "australia-southeast1",
    "asia-east1",
    "asia-northeast1",
]

columns = ["table_schema", "table_name", "column_name"]


@reacton.component
def BigQueryPanel(ds):
    data_source: ReactiveDataSource = load_data.data_source
    data_source.use()
    project_temp, set_project_temp = sol.use_state_or_update(ds.config.get("project") or data_source.value.project)

    def project():
        return ds.config.get("project") or data_source.value.project

    def get_schemas(_cancel):
        if not project():
            return None
        if data_source.value.region == "default":
            schemas_and_location_df = ds.query(f"SELECT location, schema_name FROM `{project()}`.INFORMATION_SCHEMA.SCHEMATA").to_pandas()
            default_region = schemas_and_location_df["location"].tolist()[0]
            schemas = schemas_and_location_df["schema_name"].unique().tolist()
            return SchemasResult(default_region, schemas)

        schema_df = ds.query(f"SELECT schema_name FROM `{project()}.region-{data_source.value.region}`.INFORMATION_SCHEMA.SCHEMATA").to_pandas()
        schemas = schema_df["schema_name"].unique().tolist()
        return SchemasResult(None, schemas)

    schemas_result: sol.Result = sol.hooks.use_thread(get_schemas, [data_source.value.region, project()])

    def select_single_schema():
        if schemas_result.value and len(schemas_result.value.schemas) == 1:
            value = schemas_result.value.schemas[0]
            if value != data_source.value.schema:
                data_source.set_schema(value)

    reacton.use_memo(select_single_schema, [schemas_result.value and schemas_result.value.schemas])

    def get_tables(_cancel):
        if not data_source.value.schema or not schemas_result.value:
            return

        q = f"SELECT table_id, row_count, type FROM `{project()}.{data_source.value.schema}`.__TABLES__"
        table_df = ds.query(q).to_pandas()

        return [
            {
                "text": f'{item["table_id"]} ({humanize.intword(item["row_count"] if item["type"] == 1 else "-")}) rows',
                "value": item["table_id"],
                "rows": item["row_count"],
            }
            for item in table_df.to_dict("records")
        ]

    tables_result: sol.Result = sol.hooks.use_thread(get_tables, [data_source.value.region, project(), data_source.value.schema, schemas_result.value])

    def get_column_spec(_cancel):
        if not data_source.value.schema or not schemas_result.value:
            return
        return (
            ds.query(
                f"SELECT table_name, column_name FROM `{project()}.{data_source.value.schema}`.INFORMATION_SCHEMA.COLUMNS order by table_name, ordinal_position;"  # noqa
            )
            .to_pandas()
            .groupby("table_name")["column_name"]
            .apply(list)
            .to_dict()
        )

    column_spec_result: sol.Result = sol.hooks.use_thread(get_column_spec, [data_source.value.schema, schemas_result.value])

    def check_sample():
        if data_source.value.use_query:
            data_source.set_sample(False)
            return

        if tables_result.value is None:
            return

        large_table = [item for item in tables_result.value if item["value"] == data_source.value.table and item["rows"] > 3000]
        data_source.set_sample(bool(large_table))

    reacton.use_memo(check_sample, [data_source.value.table, data_source.value.use_query])

    with v.Sheet() as main:
        with v.Row():
            with v.Col(style_="display: flex; align-items: center"):
                v.TextField(label="Project", v_model=project_temp, on_v_model=set_project_temp, disabled=bool(ds.config.get("project"))).meta(ref="project")
                if not ds.config.get("project") and project() != project_temp:
                    sol.Button(
                        "apply",
                        on_click=lambda: data_source.set_project(project_temp),
                        icon_name="mdi-check",
                        color="primary",
                        small=True,
                        class_="ml-2",
                        outlined=True,
                    ).meta(ref="apply")
            with v.Col():
                v.Select(label="region", items=regions, v_model=data_source.value.region, on_v_model=data_source.set_region).meta(ref="region")
            with v.Col():
                if schemas_result.value and schemas_result.value.schemas:
                    v.Select(label="Dataset", items=schemas_result.value.schemas, v_model=data_source.value.schema, on_v_model=data_source.set_schema).key(
                        "schema"
                    ).meta(ref="schema")
        if tables_result.value:
            with v.Row():
                with v.Col():
                    v.Switch(label="Use query").connect(Ref(data_source.fields.use_query)).meta(ref="use_query")
            with v.Row():
                with v.Col():
                    if data_source.value.use_query:
                        if column_spec_result.value:
                            SqlCode(tables=column_spec_result.value, query=data_source.value.query or "", on_query=Ref(data_source.fields.query).set).meta(
                                ref="query_editor"
                            )
                    else:
                        v.Select(label="Table", items=tables_result.value).connect(Ref(data_source.fields.table)).key("table").meta(ref="table")

        if sol.ResultState.RUNNING in [schemas_result.state, tables_result.state, column_spec_result.state]:
            v.ProgressLinear(indeterminate=True).key("progress")
        if sol.ResultState.ERROR in [schemas_result.state, tables_result.state, column_spec_result.state]:
            sol.Warning(repr(schemas_result.error or tables_result.error or column_spec_result.error))

    return main



import os
from dataclasses import replace
from typing import Any, Callable, Dict, List, Optional, cast

import mlflow
import pandas as pd
import reacton.ipyvuetify as v
import solara
from mlflow.utils.name_utils import _generate_random_name
from reacton import use_effect
from reacton.ipyvue import use_event
from solara.components.input import _use_input_numeric
from solara.lab import Ref

from domino_code_assist import util
from domino_code_assist.assistant import drawer, mixpanel
from domino_code_assist.automl.experiment_search import ExperimentSearch
from domino_code_assist.automl.store import (
    AVAILABLE_TASKS,
    MLTask,
    TaskConfig,
    get_task_arg_for_name,
    ml_task,
    task_config_store,
)
from domino_code_assist.df_select import DfSelect
from domino_code_assist.serialize import dumps
from domino_code_assist.transform.column_description import ColumnDescription

from . import flaml_api, hyper_parameters
from .select_description import SelectDescription


def bind_value(field):
    ref = Ref(field)
    return {
        "value": ref.value,
        "on_value": ref.set,
    }


def generate_code(task: MLTask):
    if task.task_type in ["Classification", "Regression", "Forecasting", "Forecasting (classification)"] and task.task_config is not None:
        with open(os.path.join(os.path.dirname(__file__), "code.py")) as f:
            code = f.read()
            template_delimiter = "# ___TEMPLATE_START___ #\n"
            code = code[code.index(template_delimiter) + len(template_delimiter) :]

            config = task.task_config
            code = code.replace("___experiment_name___", config.experiment_name or "noname")

            if task.task_type.startswith("Forecasting"):
                if config.exclude_features:
                    code = code.replace("___exclude_features___", str(config.exclude_features).replace("'", '"'))
                else:
                    code = code.replace("exclude_features = ___exclude_features___\n", "")
                filter = (
                    f'[c for c in {config.df_var_name}.columns if c not in ["{config.time_column}", *exclude_features]]'
                    if config.exclude_features
                    else f'[c for c in {config.df_var_name}.columns if c != "{config.time_column}"]'
                )
                code = code.replace(".drop(exclude_features, axis=1)", f'[["{config.time_column}"] + {filter}]')
                if util.nb_locals[config.df_var_name][config.time_column].dtypes != "datetime64[ns]":
                    code = code.replace(".copy()", f'.dropna()\ndf_automl["{config.time_column}"] = pd.to_datetime(df_automl["{config.time_column}"])\n')
                else:
                    code = code.replace(".copy()", ".dropna()")
            else:
                if not config.exclude_features:
                    code = code.replace("exclude_features = ___exclude_features___\n", "").replace("drop(exclude_features, axis=1).", "")
                else:
                    code = code.replace("___exclude_features___", str(config.exclude_features).replace("'", '"'))

            if task.task_type and task.task_type.startswith("Forecasting"):
                code = code.replace("___period___", str(config.period) if config.period is not None else "None")
            else:
                code = code.replace(", period=___period___", "")
            assert config.df_var_name is not None
            code = code.replace("___df_var_name___", config.df_var_name)

            task_arg = get_task_arg_for_name(task.task_type)

            code = code.replace("___ml_task___", task_arg)
            factor = 0
            if config.time_unit == "Minutes":
                factor = 60
            elif config.time_unit == "Hours":
                factor = 3600

            code = code.replace("___time_budget___", str(config.max_time * factor) if config.max_time is not None else "None")
            code = code.replace("___max_iterations___", str(config.max_iterations) if config.max_iterations is not None else "None")
            code = code.replace("___optimization_metric___", config.optimization_metric)
            assert config.target is not None
            code = code.replace("___label___", config.target)

            if config.auto:
                code = code.replace(", tune", "")
                code = code.replace("estimator_list = ___estimator_list___\ncustom_hp = ___custom_hp___\n\n", "")
                code = code.replace(", custom_hp=custom_hp, estimator_list=estimator_list", "")
            else:
                code = code.replace("___estimator_list___", str(config.custom_learners_enabled).replace("'", '"'))
                code = code.replace(
                    "___custom_hp___", hyper_parameters.get_hyper_parameters_string(config.custom_learners or {}, config.custom_learners_enabled or [])
                )

            if config.resampling_strategy != "Auto":
                if config.resampling_strategy == "Holdout":
                    method = "holdout"
                    arg = f"split_ratio={config.split_ratio}"
                else:
                    method = "cv"
                    arg = f"n_splits={config.folds}"
                code = code.replace(", eval_method=None", f', eval_method="{method}", {arg}')
            else:
                code = code.replace(", eval_method=None", "")
    else:
        code = f"# TODO: insert {task.task_type} code. {task.task_config}"

    return code


@solara.component
def MlTaskTile(task_info, on_task_info):
    with solara.Div(
        style_=f"display: flex; align-items: center; padding: 16px 32px {'; cursor: unset' if task_info.soon else ''}", class_="domino-ml-model-type-button"
    ).meta(ref="TaskTile") as tile:
        with solara.Div(style_="margin-right: 16px"):
            solara.Image(os.path.join(os.path.dirname(__file__), task_info.image), width="120px")
        with solara.Div():
            solara.Div(children=task_info.name, style_="font-size: 20px")
            solara.Div(children=task_info.description, style_="font-size: 14px")
            if task_info.soon:
                v.Chip(children="Coming soon", color="green", small=True, style_="margin-top: 8px", dark=True)

    def on_click(*args):
        if not task_info.soon:
            on_task_info(task_info)
        mixpanel.api.track_with_defaults("interaction", {"section": "automl", "type": "select_task", "task": task_info.name})

    v.use_event(tile, "click", on_click)


@solara.component
def FixHorizontalScrollingInDialog(children):
    # See: https://github.com/vuetifyjs/vuetify/issues/3765
    assert len(children) == 1
    use_event(children[0], "wheel.stop", lambda *args: None)
    return children[0]


@solara.component
def MlTaskConfigUI(dfs: List[str], task_type, edit: bool, on_load_own_data: Callable[[], None], emphasize_errors: bool):
    ml_task.use()
    task_config_store.use()

    expanded, set_expanded = solara.use_state([0])
    column_menu_open, set_column_menu_open = solara.use_state(cast(Optional[str], None))

    extended_dfs: List[Any] = dfs
    if not edit:
        extended_dfs = extended_dfs + [{"divider": True}, "Load more data"]

    def set_df_var_name(value):
        if value == "Load more data":
            on_load_own_data()
            return
        Ref(task_config_store.fields.df_var_name).set(value)

    df: Optional[pd.DataFrame] = util.nb_locals[task_config_store.value.df_var_name] if task_config_store.value.df_var_name else None
    available_columns: Optional[List[str]] = df.columns.tolist() if df is not None else None
    features = task_config_store.get_features(available_columns)

    def find_date_time_columns():
        if available_columns is None or not task_type.startswith("Forecasting"):
            return None

        def is_date_time(col):
            if str(col.dtypes) == "datetime64[ns]":
                return True

            if str(col.dtypes) == "object":
                try:
                    pd.to_datetime(col.head())
                    return True
                except Exception:
                    pass
            return False

        return [col for col in available_columns if is_date_time(df[col])]  # type: ignore

    date_time_columns = solara.use_memo(find_date_time_columns, [available_columns, df])

    def pre_select_time_column():
        if not date_time_columns or task_config_store.value.time_column in date_time_columns:
            return None
        Ref(task_config_store.fields.time_column).set(date_time_columns[0])

    solara.use_memo(pre_select_time_column, [date_time_columns])

    int_period, set_int_period = solara.use_state(cast(Optional[int], task_config_store.value.period))
    period, set_period, period_error = _use_input_numeric(int, optional=True, value=int_period, on_value=set_int_period)

    def set_value():
        Ref(task_config_store.fields.period).set(period if not period_error else None)

    use_effect(set_value, [period, period_error])

    def available_items(used_items: Optional[List[Optional[str]]], used_description: str):
        if available_columns is None or df is None:
            return None
        return [
            {
                "text": col + (f" ({used_description})" if col in (used_items or []) else ""),
                "value": col,
                "disabled": col in (used_items or []),
            }
            for col in available_columns
        ]

    task = get_task_arg_for_name(task_type)

    with v.Row(class_="dca-automl-config"):
        with v.Col(style_="max-width: 800px"):
            with v.ExpansionPanels(multiple=True, v_model=expanded, on_v_model=set_expanded, accordeon=True, style_="width: 800px"):
                with v.ExpansionPanel():
                    with v.ExpansionPanelHeader():
                        with solara.Div(style_="display: flex; flex-direction: row; max-width: 100%; overflow: hidden;"):
                            with solara.Div(style_="min-width: 140px"):
                                v.Html(tag="strong", children=["Data & Features "])
                            if 0 in expanded or (
                                task_config_store.value.df_var_name is None
                                and task_config_store.value.target is None
                                and task_config_store.value.exclude_features is None
                            ):
                                solara.Text("Select the data you’d like to use and value to predict").meta(ref="DataFeaturesDescription")
                            else:
                                with solara.ColumnsResponsive([6, 6, 12], gutters=False, classes=["pa-0"]):
                                    with solara.Row(gap="8px"):
                                        solara.Text("Data frame:", style="font-weight: bold; color: var(--jp-border-color1);")
                                        solara.Text(task_config_store.value.df_var_name or "")
                                    with solara.Row(gap="8px"):
                                        solara.Text("Predict value:", style="font-weight: bold; color: var(--jp-border-color1);")
                                        solara.Text(task_config_store.value.target or "")
                                    with solara.Row(gap="8px", classes=["pt-2"], style="align-items: baseline;"):
                                        solara.Text("Features:", style="font-weight: bold; color: var(--jp-border-color1);")
                                        [v.Chip(small=True, children=[col]) for col in features or []]
                                    if task_type and task_type.startswith("Forecasting"):
                                        with solara.Row(gap="8px", classes=["pt-2"], style="align-items: baseline").key("bla"):
                                            solara.Text("Period:", style="font-weight: bold; color: var(--jp-border-color1);")
                                            solara.Text(str(task_config_store.value.period) if task_config_store.value.period is not None else "")
                    with v.ExpansionPanelContent():
                        with v.Row():
                            with v.Col(class_="py-1").key("df"):
                                DfSelect(
                                    value=task_config_store.value.df_var_name,
                                    label="Data frame",
                                    on_value=set_df_var_name,
                                    items=extended_dfs or [],
                                    hide_details=bool(task_config_store.value.df_var_name),
                                    error=emphasize_errors and not bool(task_config_store.value.df_var_name),
                                    messages=["required"] if not task_config_store.value.df_var_name else None,
                                )
                            if task_type and task_type.startswith("Forecasting"):
                                with v.Col(class_="py-1").key("time"):
                                    v.Select(
                                        label="Time column",
                                        items=date_time_columns,
                                        hide_details=bool(task_config_store.value.target),
                                        v_model=task_config_store.value.time_column,
                                        on_v_model=Ref(task_config_store.fields.time_column).set,
                                        error=emphasize_errors and not bool(task_config_store.value.time_column),
                                        messages=["required"] if not task_config_store.value.time_column else None,
                                        no_data_text="No date/time columns found",
                                    ).meta(ref="TimeColumnSelect")
                            with v.Col(class_="py-1").key("predict"):
                                v.Select(
                                    label="Predict value",
                                    items=available_columns,
                                    clearable=True,
                                    hide_details=bool(task_config_store.value.target),
                                    v_model=task_config_store.value.target,
                                    on_v_model=task_config_store.set_target,
                                    error=emphasize_errors and not bool(task_config_store.value.target),
                                    messages=["required"] if not task_config_store.value.target else None,
                                )
                        with v.Row():

                            def on_features(v: List[str]):
                                Ref(task_config_store.fields.exclude_features).set(
                                    [c for c in (available_columns or []) if c not in v and c != task_config_store.value.target]
                                )

                            with v.Col(class_="py-1"):
                                v.Select(
                                    label="Features",
                                    items=available_items([task_config_store.value.target], "Predict value"),
                                    multiple=True,
                                    chips=True,
                                    deletable_chips=True,
                                    hide_details=bool(features),
                                    error=emphasize_errors and not bool(features),
                                    messages=["required"] if not features else None,
                                    v_model=features,
                                    on_v_model=on_features,
                                )
                        if task_type and task_type.startswith("Forecasting"):
                            with v.Row():
                                with v.Col(class_="py-1", sm=2):
                                    period_error_message = None
                                    if period is None:
                                        period_error_message = "required"
                                    elif period_error:
                                        period_error_message = "invalid"

                                    v.TextField(
                                        label="Period",
                                        clearable=True,
                                        v_model=period,
                                        on_v_model=set_period,
                                        error=emphasize_errors and bool(period_error_message),
                                        messages=period_error_message,
                                        hide_details=bool(not period_error_message),
                                    )
                                with v.Col(class_="py-1", sm=2):
                                    with solara.ToggleButtonsSingle(
                                        value="regression" if ml_task.value.task_type == "Forecasting" else "classification",  # type: ignore
                                        on_value=lambda val: Ref(ml_task.fields.task_type).set(
                                            "Forecasting" if val == "regression" else "Forecasting (classification)"
                                        ),
                                    ):
                                        solara.Button("Regression", value="regression", small=True, classes=["mt-2"])
                                        solara.Button("Classification", value="classification", small=True, classes=["mt-2"])
                with v.ExpansionPanel():
                    with v.ExpansionPanelHeader():
                        with solara.Div(style_="display: flex; flex-direction: row;"):
                            with solara.Div(style_="min-width: 140px"):
                                v.Html(tag="strong", children=["Training"])
                            if 1 in expanded or (
                                task_config_store.value.experiment_name is None
                                and task_config_store.value.optimization_metric == "auto"
                                and task_config_store.value.max_iterations is None
                                and task_config_store.value.max_time is None
                            ):
                                solara.Text("Choose the metric to optimize and your time budget")
                            else:
                                with solara.ColumnsResponsive([12, 4, 4, 4], gutters=False, classes=["pa-0"]):
                                    with solara.Row(gap="8px"):
                                        solara.Text("Experiment name:", style="font-weight: bold; color: var(--jp-border-color1);")
                                        solara.Text(task_config_store.value.experiment_name or "")
                                    with solara.Row(gap="8px", classes=["pt-2"]):
                                        solara.Text("Optimize metric:", style="font-weight: bold; color: var(--jp-border-color1);")
                                        solara.Text(task_config_store.value.optimization_metric or "")
                                    with solara.Row(gap="8px", classes=["pt-2"]):
                                        solara.Text("Max iterations:", style="font-weight: bold; color: var(--jp-border-color1);")
                                        solara.Text(str(task_config_store.value.max_iterations) if task_config_store.value.max_iterations else "")
                                    with solara.Row(gap="8px", classes=["pt-2"]):
                                        solara.Text("Time budget:", style="font-weight: bold; color: var(--jp-border-color1);")
                                        solara.Text(
                                            (str(task_config_store.value.max_time) if task_config_store.value.max_time else "")
                                            + " "
                                            + task_config_store.value.time_unit
                                        ) if task_config_store.value.max_time else ""

                    with v.ExpansionPanelContent():
                        with v.Row():
                            with v.Col(class_="py-1"):
                                ExperimentSearch(**bind_value(task_config_store.fields.experiment_name))
                        with v.Row():
                            with v.Col(class_="py-1"):
                                optimization_metric_options = [
                                    {
                                        "value": key,
                                        "text": key,
                                        "description": description,
                                    }
                                    for key, description in flaml_api.common_optimization_metric_options
                                    if task_config_store.value.optimization_metric == "auto" or key != "auto"
                                ]
                                SelectDescription(
                                    label="Optimize metric", items=optimization_metric_options, **bind_value(task_config_store.fields.optimization_metric)
                                )
                            with v.Col(class_="py-1"):
                                solara.InputInt(
                                    label="Max iterations",
                                    continuous_update=True,
                                    clearable=True,
                                    optional=True,
                                    **bind_value(task_config_store.fields.max_iterations),
                                )
                            with v.Col(class_="py-1"):
                                solara.InputInt(
                                    label="Time budget",
                                    continuous_update=True,
                                    clearable=True,
                                    optional=True,
                                    **bind_value(task_config_store.fields.max_time),
                                )
                            with v.Col(class_="py-1"):
                                v.Select(
                                    label="Unit",
                                    items=["Seconds", "Minutes", "Hours"],
                                    hide_details=True,
                                ).connect(Ref(task_config_store.fields.time_unit))
                hyper_parameters.Hyperparameters(task)
                with v.ExpansionPanel():
                    with v.ExpansionPanelHeader():
                        with solara.Div(style_="display: flex; flex-direction: row;"):
                            with solara.Div(style_="min-width: 140px"):
                                v.Html(tag="strong", children=["Resampling"])
                            if 3 in expanded or (task_config_store.value.resampling_strategy == "Auto"):
                                solara.Text("Customize the resampling method")
                            else:
                                with solara.ColumnsResponsive([6, 6], gutters=False, classes=["pa-0"]):
                                    with solara.Row(gap="8px"):
                                        solara.Text("Resampling strategy:", style="font-weight: bold; color: var(--jp-border-color1);")
                                        solara.Text(task_config_store.value.resampling_strategy)
                                    with solara.Row(gap="8px", classes=["pt-2"]):
                                        if task_config_store.value.resampling_strategy == "Holdout":
                                            solara.Text(
                                                "Split ratio:",
                                                style="font-weight: bold; color: var(--jp-border-color1);",
                                            )
                                            solara.Text(task_config_store.value.split_ratio)
                                        else:
                                            solara.Text(
                                                "Folds:",
                                                style="font-weight: bold; color: var(--jp-border-color1);",
                                            )
                                            solara.Text(task_config_store.value.folds)
                    with v.ExpansionPanelContent():
                        with v.Row():
                            with v.Col(class_="py-1"):
                                v.Select(label="Validation method", items=["Auto", "Holdout", "Cross validation"]).connect(
                                    Ref(task_config_store.fields.resampling_strategy)
                                )
                            with v.Col(class_="py-1"):
                                if task_config_store.value.resampling_strategy == "Holdout":
                                    messages = None
                                    if not task_config_store.value.split_ratio:
                                        messages = ["required"]
                                    elif util.to_float(task_config_store.value.split_ratio) is None:
                                        messages = ["invalid"]
                                    v.TextField(label="Split ratio", error_messages=messages).connect(Ref(task_config_store.fields.split_ratio)).key(
                                        "split_ratio"
                                    )
                                elif task_config_store.value.resampling_strategy == "Cross validation":
                                    messages = None
                                    if not task_config_store.value.folds:
                                        messages = ["required"]
                                    elif util.to_int(task_config_store.value.folds) is None:
                                        messages = ["invalid"]
                                    v.TextField(label="Folds", error=emphasize_errors and bool(messages), messages=messages).key("folds").connect(
                                        Ref(task_config_store.fields.folds)
                                    )
        with FixHorizontalScrollingInDialog():
            with v.Col(style_="overflow: hidden;"):
                if df is not None:
                    solara.DataFrame(
                        df,
                        items_per_page=10,
                        scrollable=True,
                        on_column_header_hover=set_column_menu_open,
                        column_header_info=ColumnDescription(df, column_menu_open),
                    )


# For remembering params when switching between forecasting tasks
hyper_param_scratchpad = solara.reactive(cast(Dict[str, Any], {}))


@solara.component
def AutoMlDrawer(
    edit_data: Optional[MLTask],
    set_code: Callable[[Dict], None],
    dfs: List[str],
    is_open: bool,
    edit: bool,
    on_close: Callable[[], None],
    on_load_own_data: Callable[[], None],
    overwrite_warning=None,
):
    ml_task.use()
    task_config_store.use()

    emphasize_errors, set_emphasize_errors = solara.use_state(False)

    def is_valid():
        return (
            ml_task.value.task_type in ["Regression", "Classification"]
            or (
                ml_task.value.task_type
                and ml_task.value.task_type.startswith("Forecasting")
                and task_config_store.value.period is not None
                and task_config_store.value.time_column is not None
            )
        ) and task_config_store.is_valid()

    def on_apply():
        if is_valid():
            complete_task = ml_task.value
            if complete_task.task_type in ["Classification", "Regression", "Forecasting", "Forecasting (classification)"]:
                task_config = replace(
                    task_config_store.value,
                    auto=hyper_parameters.auto.value,
                    custom_learners=hyper_parameters.learner_params.value,
                    custom_learners_enabled=hyper_parameters.selected_learners.value,
                )
                complete_task = replace(complete_task, task_config=task_config)
            set_code(
                {
                    "code": generate_code(complete_task),
                    "meta": dumps(complete_task),
                }
            )
            mixpanel.api.track_with_defaults("inserted code", {"section": "automl", "task_type": complete_task.task_type})
            on_close()
        else:
            set_emphasize_errors(True)

    def init():
        hyper_param_scratchpad.set({})
        set_emphasize_errors(False)
        if is_open and edit_data and edit:
            ml_task.set(edit_data)
            assert edit_data.task_config
            task_config_store.set(edit_data.task_config)

            assert edit_data.task_config
            hyper_parameters.auto.value = edit_data.task_config.auto
            hyper_parameters.learner_params.value = edit_data.task_config.custom_learners or {}
            hyper_parameters.selected_learners.value = edit_data.task_config.custom_learners_enabled or []
        else:
            ml_task.set(MLTask())
            try:
                experiment_name = mlflow.get_experiment("0").name
            except Exception:
                experiment_name = _generate_random_name()
            task_config_store.set(
                TaskConfig(df_var_name=dfs[0], experiment_name=experiment_name, max_iterations=50, max_time=5)
                if dfs
                else TaskConfig(experiment_name=experiment_name, max_iterations=50, max_time=5)
            )

            hyper_parameters.auto.value = True
            hyper_parameters.learner_params.value = {}
            hyper_parameters.selected_learners.value = None

    solara.use_memo(init, [is_open])

    last_task_type = solara.use_previous(ml_task.value.task_type, [ml_task.value.task_type])

    def on_ml_task():
        current = hyper_param_scratchpad.value
        if last_task_type:
            hyper_param_scratchpad.set(
                {
                    "auto": hyper_parameters.auto.value,
                    "learner_params": hyper_parameters.learner_params.value,
                    "selected_learners": hyper_parameters.selected_learners.value,
                }
            )
            if current:
                hyper_parameters.auto.value = current["auto"]
                hyper_parameters.learner_params.value = current["learner_params"]
                hyper_parameters.selected_learners.value = current["selected_learners"]
            else:
                hyper_parameters.auto.value = True
                hyper_parameters.learner_params.value = {}
                hyper_parameters.selected_learners.value = flaml_api.get_estimator_ids_for_task(get_task_arg_for_name(ml_task.value.task_type))

        if hyper_parameters.selected_learners.value is None and ml_task.value.task_type:
            hyper_parameters.selected_learners.value = flaml_api.get_estimator_ids_for_task(get_task_arg_for_name(ml_task.value.task_type))

    solara.use_effect(on_ml_task, [ml_task.value.task_type])

    def load_own_data():
        on_close()
        on_load_own_data()

    with drawer.RightDrawer(
        open=is_open,
        on_open=lambda v: on_close() if not v else None,
        title="AutoML" + (" - " + ml_task.value.task_type if ml_task.value.task_type else ""),
        edit=bool(edit),
        on_apply=on_apply,
        apply_disabled=not ml_task.value.task_type,
        show_var_out=False,
        width="1024px" if not ml_task.value.task_type else "100vw",
        warning_widget=overwrite_warning,
        class_="dca-ml-task-selection",
    ):
        with v.Stepper(v_model=1 if not ml_task.value.task_type else 2, elevation=0, class_="elevation-0", style_="margin: 0"):
            with v.StepperItems():
                with v.StepperContent(step=1):
                    v.Html(
                        tag="p",
                        style_="background-color: var(--jp-border-color3); padding:32px;",
                        children=[
                            "Choose what type of machine learning task you’d like to use with AutoML. ",
                            "For more on how to leverage the power of AutoML in Domino, ",
                            v.Html(
                                tag="a",
                                attributes={"href": "https://docs.dominodatalab.com/en/latest/user_guide/795e6f/domino-code-assist/", "target": "_blank"},
                                children=["check out the docs here"],
                            ),
                            ".",
                        ],
                    )
                    with solara.Div():
                        for task in [t for t in AVAILABLE_TASKS if t.name != "Forecasting (classification)"]:
                            MlTaskTile(task, on_task_info=lambda task_info: Ref(ml_task.fields.task_type).set(task_info.name) if task_info else None)
                with v.StepperContent(step=2):
                    v.Html(
                        tag="p",
                        style_="background-color: var(--jp-border-color3); padding: 32px; margin-bottom: 0;",
                        children=[
                            "Now configure the settings for how you’d like to run the AutoML. For more on how to leverage the power of AutoML in Domino, ",
                            v.Html(
                                tag="a",
                                attributes={"href": "https://docs.dominodatalab.com/en/latest/user_guide/795e6f/domino-code-assist/", "target": "_blank"},
                                children=["check out the docs here"],
                            ),
                            ".",
                        ],
                    )
                    with solara.Div(style_="padding: 32px"):
                        if ml_task.value.task_type:
                            if ml_task.value.task_type in ["Classification", "Regression", "Forecasting", "Forecasting (classification)"]:
                                MlTaskConfigUI(dfs, ml_task.value.task_type, edit, load_own_data, emphasize_errors)
                            else:
                                v.Chip(children=[ml_task.value.task_type])
